import pandas as pd
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer
# from sklearn.linear_model import LogisticRegression, Lasso
from nlpia.constants import DATA_DIR

df = pd.read_csv(DATA_DIR / 'baby-names-region.csv.gz')

# df = df.sample(1_000_000, random_state=1989)
np.random.seed(451)
istrain = np.random.rand(len(df)) < .9

vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
vectorizer.fit(df['name'][istrain])
vecs = vectorizer.transform(df['name'])



import pandas as pd
pd.options.display.width = 120  # <1>

DATA_DIR = ('https://gitlab.com/tangibleai/nlpia/-/raw/master/src/nlpia/data')
url= DATA_DIR + '/sms-spam.csv'
sms = pd.read_csv(url)
index = ['sms{}{}'.format(i, '!'*j) for (i,j) in zip(range(len(sms)), sms.spam)]  # <2>
sms = pd.DataFrame(sms.values, columns=sms.columns, index=index)
mask = sms.spam.astype(bool).values
sms['spam'] = sms.spam.astype(int)
"""
>>> sms.head(6)
       spam                                               text
sms0      0  Go until jurong point, crazy.. Available only ...
sms1      0                      Ok lar... Joking wif u oni...
sms2!     1  Free entry in 2 a wkly comp to win FA Cup fina...
sms3      0  U dun say so early hor... U c already then say...
sms4      0  Nah I don't think he goes to usf, he lives aro...
sms5!     1  FreeMsg Hey there darling it's been 3 week's n...
"""

from sklearn.feature_extraction.text import TfidfVectorizer
import spacy

nlp = spacy.load("en_core_web_sm")

def spacy_tokenizer(sentence):
    return [token.text for token in nlp(sentence.lower())]

tfidf_model = TfidfVectorizer(tokenizer=spacy_tokenizer)
tfidf_docs = tfidf_model.fit_transform(raw_documents=sms.text).toarray()
"""
>>> from sklearn.feature_extraction.text import TfidfVectorizer
>>> from nltk.tokenize.casual import casual_tokenize
>>> tfidf_model = TfidfVectorizer(tokenizer=casual_tokenize)
>>> tfidf_docs = tfidf_model.fit_transform(raw_documents=sms.text).toarray()
>>> tfidf_docs.shape
(4837, 9232)
>>> sms.spam.sum()
638
"""

mask = sms.spam.astype(bool).values  # <1>
spam_centroid = tfidf_docs[mask].mean(axis=0) # <2>
ham_centroid = tfidf_docs[~mask].mean(axis=0)
"""
>>> mask = sms.spam.astype(bool)
>>> spam_centroid = tfidf_docs[mask].mean(axis=0)
>>> spam_centroid.round(2)
array([0.06, 0.  , 0.  , ..., 0.  , 0.  , 0.  ])
>>> ham_centroid = tfidf_docs[~mask].mean(axis=0)
>>> ham_centroid.round(2)
array([0.02, 0.01, 0.  , ..., 0.  , 0.  , 0.  ])
"""

spamminess_score = tfidf_docs.dot(spam_centroid - ham_centroid)
"""
>>> spamminess_score = tfidf_docs.dot(spam_centroid - ham_centroid)
>>> spamminess_score
array([-0.01469806, -0.02007376,  0.03856095, ..., -0.01014774, -0.00344281,  0.00395752])
"""

from sklearn.preprocessing import MinMaxScaler
sms['lda_score'] = MinMaxScaler().fit_transform(spamminess_score.reshape(-1,1))
sms['lda_predict'] = (sms.lda_score > .5).astype(int)

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
lda_tfidf = LinearDiscriminantAnalysis()
lda_tfidf = lda_tfidf.fit(tfidf_docs, sms['spam'])  # <1>
sms['lda_tfidf_predict'] = lda_tfidf.predict(tfidf_docs)
round(float(lda_tfidf.score(tfidf_docs, sms['spam'])), 3)

"""
>>> from sklearn.preprocessing import MinMaxScaler
>>> sms['lda_score'] = MinMaxScaler().fit_transform(spamminess_score.reshape(-1,1))
>>> sms['lda_predict'] = (sms.lda_score > .5).astype(int)
>>> sms['spam lda_predict lda_score'.split()].round(2).head(6)
       spam  lda_predict  lda_score
sms0      0            0       0.23
sms1      0            0       0.18
sms2!     1            1       0.72
sms3      0            0       0.18
sms4      0            0       0.29
sms5!     1            1       0.55
"""
from sklearn.model_selection import train_test_split
X_train, X_test, y_train, y_test = train_test_split(tfidf_docs,\
     sms.spam.values, test_size=0.5, random_state=271828)
lda_tfidf = lda_tfidf.fit(X_train, y_train)  # <2>
round(float(lda_tfidf.score(X_train, y_train)), 3)
round(float(lda_tfidf.score(X_test, y_test)), 3)

# running Hobson's code from previous book

from nltk.tokenize.casual import casual_tokenize
import spacy

nlp = spacy.load("en_core_web_sm")

def spacy_tokenizer(sentence):
    return [token.text for token in nlp(sentence.lower())]

tfidf = TfidfVectorizer(tokenizer=spacy_tokenizer)
tfidf_docs = tfidf.fit_transform(raw_documents=sms.text).toarray()
tfidf_docs = tfidf_docs - tfidf_docs.mean(axis=0)

X_train, X_test, y_train, y_test = train_test_split(tfidf_docs, sms.spam.values, test_size = 0.5, random_state = 271828)
lda = LinearDiscriminantAnalysis(n_components=1)
lda = lda.fit(X_train, y_train)



from sklearn.decomposition import PCA
from matplotlib import pyplot as plt
import seaborn
pca_model = PCA(n_components=3)
tfidf_docs_3d = pca_model.fit_transform(tfidf_docs)
df = pd.DataFrame(tfidf_docs_3d)
ax = df[~mask].plot(x=0, y=1, kind='scatter', alpha=.5, c='green')
df[mask].plot(x=0, y=1, ax=ax, alpha=.1, kind='scatter', c='red')
plt.xlabel(' x')
plt.ylabel(' y')
plt.show()

"""
from sklearn.decomposition import PCA
from matplotlib import pyplot as plt
import seaborn
pca_model = PCA(n_components=3)
tfidf_docs_3d = pca_model.fit_transform(tfidf_docs)
df = pd.DataFrame(tfidf_docs_3d)

import plotly as py
spam_trace = dict(
        x=df[0][mask], y=df[1][mask], z=df[2][mask],
        type="scatter3d", mode='markers',
        marker= dict(size=3, color='red', line=dict(width=0))
    )
ham_trace = dict(
        x=df[0][~mask], y=df[1][~mask], z=df[2][~mask],
        type="scatter3d", mode='markers',
        marker= dict(size=3, color='green', line=dict(width=0))
    )
fig = dict(data=[ham_trace, spam_trace], layout={'title': 'LDA Spamminess Model'})
py.offline.plot(fig, filename='lda_toxicity_3d_scatter.html')
"""


# from: https://medium.com/biaslyai/pytorch-linear-and-logistic-regression-models-5c5f0da2cb9
import torch
from torch.autograd import Variable
# from torch.nn import functional as F
from torch import nn


x_data = Variable(torch.Tensor([[10.0], [9.0], [3.0], [2.0]]))
y_data = Variable(torch.Tensor([[90.0], [80.0], [50.0], [30.0]]))


class LinearRegression(torch.nn.Module):
    def __init__(self, num_inputs=1, num_neurons=1):
        super().__init__()
        self.linear = nn.Linear(
            in_features=num_inputs,
            out_features=num_neurons)

    def forward(self, x):
        y_pred = self.linear(x)
        return y_pred


model = LinearRegression()

"""
Slopes and intercepts are initialized to random values between 0 and 1.

>>> torch.manual_seed(3200)
>>> model=LinearRegression()
>>> list(model.parameters())
>>> [Parameter containing:
 tensor([[0.2490]], requires_grad=True),
 Parameter containing:
 tensor([0.7142], requires_grad=True)]

>>> model = LinearRegression()
>>> list(model.parameters())
[Parameter containing:
 tensor([[0.9969]], requires_grad=True),
 Parameter containing:
 tensor([-0.9100], requires_grad=True)]

"""

# lr = learning rate
# SGD(params, lr=<required parameter>, momentum=0, dampening=0, weight_decay=0, nesterov=False)
optimizer = torch.optim.SGD(params=model.parameters(), lr=0.01)

# Adam(params, lr=0.001, betas=(0.9, 0.999), eps=1e-08, weight_decay=0, amsgrad=False)

# MSELoss(size_average=None, reduce=None, reduction: str = 'mean')
criterion = torch.nn.MSELoss(size_average=False)


for epoch in range(20):
    model.train()
    optimizer.zero_grad()

    y_pred = model(x_data)
    loss = criterion(y_pred, y_data)

    loss.backward()
    optimizer.step()



"""
References:
- Aransil's Github: (https://github.com/redransil/plotly-dirgraph)
- Roland Smith (https://stackoverflow.com/a/59538246/623735)
- Ploty Docs: (https://plotly.com/python/network-graphs/)

"""

import math
from typing import List
from itertools import chain

# Start and end are lists defining start and end points
# Edge x and y are lists used to construct the graph
# arrowAngle and arrowLength define properties of the arrowhead
# arrowPos is None, 'middle' or 'end' based on where on the edge you want the arrow to appear
# arrowLength is the length of the arrowhead
# arrowAngle is the angle in degrees that the arrowhead makes with the edge
# dotSize is the plotly scatter dot size you are using (used to even out line spacing when you have a mix of edge lengths)
def addEdge(start, end, edge_x, edge_y, lengthFrac=1, arrowPos=None, arrowLength=0.025, arrowAngle=30, dotSize=20):

    # Get start and end cartesian coordinates
    x0, y0 = start
    x1, y1 = end

    # Incorporate the fraction of this segment covered by a dot into total reduction
    length = math.sqrt((x1 - x0)**2 + (y1 - y0)**2)
    dotSizeConversion = .0565 / 20  # length units per dot size
    convertedDotDiameter = dotSize * dotSizeConversion
    lengthFracReduction = convertedDotDiameter / length
    lengthFrac = lengthFrac - lengthFracReduction

    # If the line segment should not cover the entire distance, get actual start and end coords
    skipX = (x1 - x0) * (1 - lengthFrac)
    skipY = (y1 - y0) * (1 - lengthFrac)
    x0 = x0 + skipX / 2
    x1 = x1 - skipX / 2
    y0 = y0 + skipY / 2
    y1 = y1 - skipY / 2

    # Append line corresponding to the edge
    edge_x.append(x0)
    edge_x.append(x1)
    edge_x.append(None)  # Prevents a line being drawn from end of this edge to start of next edge
    edge_y.append(y0)
    edge_y.append(y1)
    edge_y.append(None)

    # Draw arrow
    if not arrowPos == None:

        # Find the point of the arrow; assume is at end unless told middle
        pointx = x1
        pointy = y1

        eta = math.degrees(math.atan((x1 - x0) / (y1 - y0))) if y1 != y0 else 90.0

        if arrowPos == 'middle' or arrowPos == 'mid':
            pointx = x0 + (x1 - x0) / 2
            pointy = y0 + (y1 - y0) / 2

        # Find the directions the arrows are pointing
        signx = (x1 - x0) / abs(x1 - x0) if x1 != x0 else +1  # verify this once
        signy = (y1 - y0) / abs(y1 - y0) if y1 != y0 else +1  # verified

        # Append first arrowhead
        dx = arrowLength * math.sin(math.radians(eta + arrowAngle))
        dy = arrowLength * math.cos(math.radians(eta + arrowAngle))
        edge_x.append(pointx)
        edge_x.append(pointx - signx**2 * signy * dx)
        edge_x.append(None)
        edge_y.append(pointy)
        edge_y.append(pointy - signx**2 * signy * dy)
        edge_y.append(None)

        # And second arrowhead
        dx = arrowLength * math.sin(math.radians(eta - arrowAngle))
        dy = arrowLength * math.cos(math.radians(eta - arrowAngle))
        edge_x.append(pointx)
        edge_x.append(pointx - signx**2 * signy * dx)
        edge_x.append(None)
        edge_y.append(pointy)
        edge_y.append(pointy - signx**2 * signy * dy)
        edge_y.append(None)

    return edge_x, edge_y


def add_arrows(source_x: List[float], target_x: List[float], source_y: List[float], target_y: List[float],
               arrowLength=0.025, arrowAngle=30):
    pointx = list(map(lambda x: x[0] + (x[1] - x[0]) / 2, zip(source_x, target_x)))
    pointy = list(map(lambda x: x[0] + (x[1] - x[0]) / 2, zip(source_y, target_y)))
    etas = list(map(lambda x: math.degrees(math.atan((x[1] - x[0]) / (x[3] - x[2]))),
                    zip(source_x, target_x, source_y, target_y)))

    signx = list(map(lambda x: (x[1] - x[0]) / abs(x[1] - x[0]), zip(source_x, target_x)))
    signy = list(map(lambda x: (x[1] - x[0]) / abs(x[1] - x[0]), zip(source_y, target_y)))

    dx = list(map(lambda x: arrowLength * math.sin(math.radians(x + arrowAngle)), etas))
    dy = list(map(lambda x: arrowLength * math.cos(math.radians(x + arrowAngle)), etas))
    none_spacer = [None for _ in range(len(pointx))]
    arrow_line_x = list(map(lambda x: x[0] - x[1] ** 2 * x[2] * x[3], zip(pointx, signx, signy, dx)))
    arrow_line_y = list(map(lambda x: x[0] - x[1] ** 2 * x[2] * x[3], zip(pointy, signx, signy, dy)))

    arrow_line_1x_coords = list(chain(*zip(pointx, arrow_line_x, none_spacer)))
    arrow_line_1y_coords = list(chain(*zip(pointy, arrow_line_y, none_spacer)))

    dx = list(map(lambda x: arrowLength * math.sin(math.radians(x - arrowAngle)), etas))
    dy = list(map(lambda x: arrowLength * math.cos(math.radians(x - arrowAngle)), etas))
    none_spacer = [None for _ in range(len(pointx))]
    arrow_line_x = list(map(lambda x: x[0] - x[1] ** 2 * x[2] * x[3], zip(pointx, signx, signy, dx)))
    arrow_line_y = list(map(lambda x: x[0] - x[1] ** 2 * x[2] * x[3], zip(pointy, signx, signy, dy)))

    arrow_line_2x_coords = list(chain(*zip(pointx, arrow_line_x, none_spacer)))
    arrow_line_2y_coords = list(chain(*zip(pointy, arrow_line_y, none_spacer)))

    x_arrows = arrow_line_1x_coords + arrow_line_2x_coords
    y_arrows = arrow_line_1y_coords + arrow_line_2y_coords

    return x_arrows, y_arrows



import numpy as np
import pandas as pd
# neither year nor len are statistically significant predictors of sex
from pathlib import Path
import numpy as np
import pandas as pd
DATA_DIR = Path('.nlpia2-data')
df = pd.read_csv(DATA_DIR / 'baby-names-region.csv.gz')

np.random.seed(451)
df = df.sample(10_000)
names = df['name'].unique()
istrain = np.random.rand(len(names)) < .9
istest = ~istrain
df.groupby('name')['istrain'] = istrain
groups = df.groupby('name')
dfistrain = pd.DataFrame([istrain]).T
dfistrain
dfistrain.columns = ['istrain']
sum(dfistrain)
dfistrain.sum()
dfistrain.index = names
df.merge(dfistrain, on='name', kind='left')
df.merge?
df.merge?
df.merge(dfistrain, on='name')
dfistrain?
dfistrain
dfistrain['name'] = names
df.merge(dfistrain, on='name')
hist

istrain = df.merge(dfistrain, on='name')['istrain']
istrain
from sklearn.linear_model import LogisticRegression
model = LogisticRegression()
model.fit(df['name'].str.len()[istrain], df['sex'])
istrain = istrain.values
model.fit(df['name'].str.len()[istrain], df['sex'])
model.fit(df[['name']].str.len()[istrain], df['sex'])
X = df['name'].str.len()
df['len'] = df['name'].str.len()
X = df[['len']]
y = df['sex']
model.fit(X[istrain], y[istrain])
mmodel.score(X[istrain], y[istrain])
model.score(X[istrain], y[istrain])
model.score(X[~istrain], y[~istrain])
history -o -p
model.score(X[istrain], y[istrain], sample_weight=df['count'])
model.score(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model.fit(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model.score(X[istrain], y[istrain])
model.score(X[~istrain], y~[istrain])
model.score(X[~istrain], y[~istrain])
model.score(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model.score(X[~istrain], y[~istrain], sample_weight=df['count'][~istrain])
hist -o -p
history
hist -o -p
vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
vectorizer.fit(names.index.values[istrain])
vecs = vectorizer.transform(names.index)
from sklearn.feature_extraction.text TfidfVectorizer
from sklearn.feature_extraction.text import TfidfVectorizer
vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
vectorizer.fit(names.index.values[istrain])
vecs = vectorizer.transform(names.index)
vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
vectorizer.fit(df['name'][istrain])
vectorizer
vecs = vectorizer.transform(df['names'])
vecs = vectorizer.transform(df['name'])
vecs[istrain]
dfvecs = pd.DataFrame(vecs, dtype=pd.SparseDtype("float"))
pd.SparseDataFrame
dfvecs = pd.DataFrame.sparse.from_spmatrix(vecs)
dfvecs.columns = vectorizer.get_feature_names_out()
dfvecs
dfvecs.head(2).T.head(100)
vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3), lowercase=False)
vectorizer.fit(df['name'][istrain])
vectorizer
vecs = vectorizer.transform(df['name'])
dfvecs = pd.DataFrame.sparse.from_spmatrix(vecs)
dfvecs.head(2).T.head(100)
dfvecs.columns = vectorizer.get_feature_names_out()
dfvecs.head(5).T.head(100)
df['name'].head()
dfvecs.index = names
dfvecs.index = df['name']
dfvecs.head(5).T.head(100)
dfvecs['Amy'].head(5).T.head(100)
dfvecs[['Amy']].T.head(100)
dfvecs[['Amy']]
dfvecs
dfvecs.head()
dfvecs.head().T
dfvecs.T[['Amy', 'Alex']]
dfvecs.T[['Amy', 'Alex']].head().T
hist
X = dfvecs
model = LogisticRegression()
model.fit(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model = LogisticRegression(max_iterations=5000)
model = LogisticRegression(C=.5, max_iterations=5000)
model = LogisticRegression(C=.5, max_iter=5000)
model.fit(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model.score(X[~istrain], y[~istrain], sample_weight=df['count'][~istrain])
model.score(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model = LogisticRegression(C=10, max_iter=5000)
model.fit(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model.score(X[~istrain], y[~istrain], sample_weight=df['count'][~istrain])
model.fit(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model.fit(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model.score(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model.score(X[~istrain], y[~istrain], sample_weight=df['count'][~istrain])
hist -o -p
model.classes_
model.predict_proba(vectorizer.transform(['Kemal', 'Syndee', 'Hobson', 'Maria']))
pd.DataFrame(model.predict_proba(vectorizer.transform(['Kemal', 'Syndee', 'Hobson', 'Maria'])), columns = model.classes_)
pd.DataFrame(model.predict_proba(vectorizer.transform(['Kemal', 'Syndee', 'Hobson', 'Maria'])),columns = model.classes_)
pd.DataFrame(model.predict_proba(vectorizer.transform(['Kemal', 'Syndee', 'Hobson', 'Maria'])),
columns = model.classes_, index=['Kemal', 'Syndee', 'Hobson', 'Maria'])
'Kemal' in names
'Syndee' in names
'Hobson' in names
'Maria' in names
'Maria' in names[istrain]
'Maria' in df['name'][istrain].unique()
hist
df['name'][istrain].unique()
ournames = ['Kemal', 'Syndee', 'Hobson', 'Maria']
probas = model.predict_proba(vectorizer.transform(ournames))
df_probas = pd.DataFrame(probas, index=ournames, columns=model.classes_)

training_set_names = df['name'][istrain].unique()
for n in ournames:
    print(n, n in training_set_names)
vectorizer.transform(ournames)
model.predict_proba(vectorizer.transform(ournames))
model.predict_proba(vectorizer.transform(ournames))
model.predict_proba(vectorizer.transform(ournames))
model.predict_proba(vectorizer.transform(ournames), feature_names=vectorizer.get_feature_names_out())
model.feature_names_in_
X_ournames = vectorizer.transform(ournames)
X.shape
X_ournames.shape
type(X)
type(X_ournames)
X_ournames = pd.DataFrame.sparse.from_spmatrix(vectorizer.transform(ournames), index=ournames, columns=vectorizer.get_feature_names_in())
X_ournames = pd.DataFrame.sparse.from_spmatrix(vectorizer.transform(ournames), index=ournames, columns=vectorizer.get_feature_names_out())
X_ournames
model.predict_proba(X_ournames)
hist
hist -o -p
pd.DataFrame(model.predict_proba(X_ournames), columns=model.classes_, index=ournames)
hist -o -p
pd.DataFrame(model.predict_proba(X_ournames), 
    columns=model.classes_,
    index=ournames).round(3)
pd.DataFrame(model.predict_proba(X_ournames), 
    columns=model.classes_,
    index=ournames).round(4)
hist -o -p
hist -o -p -f kemal_syndee_unique_names_no_cheating_sample_weight.md
hist -f kemal_syndee_unique_names_no_cheating_sample_weight.py



import datasets
datasets.search('TREC')
from datasets import *
who
dir(search)
vars(search)
dir(search)
who
dataset_dict
datasets?
help(datasets)
tasks
dir(tasks)
dir(tasks.TextClassification)
help(tasks.TextClassification)
trec = load_dataset('trec')
trecqa = load_dataset('trec-qa')
trecqa = load_dataset('trec-qc')
trecqa = load_dataset('SetFit/TREC-QC')
trecqc = trecqa
dir(trecqc)
dir(trecqc.data)
len(trecqc.data)
lentrecqc.data
trecqc.data
trecqc.data[0]
trecqc.data.keys()
len(trecqc.data['test'])
len(trecqc.data['train'])
len(trec.data['train'])
len(trec.data['test'])
df = trec.data['train']
df.shape
df.columns
df = df.to_pandas()
df.columns
trec
vars(trec)
trec.files
list_datasets()
names = _
[n for n in names if 'trec' in n.lower()]
[n for n in names if 'qa' in n.lower()]
[n for n in names if 'QA' in n]
trec.info
load_dataset?
trec = load_dataset('trec', save_infos=True)
trec.info
trec.keys()
trec['train'].info
print(trec['train'].info)
print(trec['train'].info.description)
print(trec['train'].info.license)
vars(trec['train'].info)
dir(trec['train'].info)
print(trec['train'].info.features)
print(trec['train'].info.features.label-couarse)
print(trec['train'].info.features[label-coarse])
print(trec['train'].info.features['label-coarse'])
print(trec['train'].info.features['label-fine'])
df = pd.concat(trec[k].data.to_pandas() for k in trec[k])
import pandas as pd
df = pd.concat(trec[k].data.to_pandas() for k in trec[k], axis=0)
df = pd.concat([trec[k].data.to_pandas() for k in trec[k]], axis=0)
df = pd.concat([trec[k].data.to_pandas() for k in trec], axis=0)
df.shape
trec['train'].data.shape
trec = load_dataset('trec', split='all')
trec.data.shape
trec = load_dataset('trec')
trec.data.shape
trec.keys()
trec = load_dataset('trec', split='all')
df = trec.data.to_pandas()
df.shape
dir(df.data)
dir(trec.data)
dir(trec.data.features
)
trec.data.features
trec.features
trec.features['label_coarse']
labels_course = trec.features['label-coarse']
labels_fine = trec.features['label-fine']
labels_fine
labels_course = list(trec.features['label-coarse'])
labels_course = list(trec.features['label-coarse'].names)
labels_fine = trec.features['label-fine'].names
labels_fine
labels_course = trec.features['label-coarse'].names
id2name = dict(enumerate(labels_course))
labels_coarse = list(trec.features['label-coarse'].names)
labels_fine = list(trec.features['label-fine'].names)
df.shape
df.to_csv('/home/hobs/code/nlpia2/src/nlpia2/ch08/cnn-classify-text/data/trec.csv')
df.to_csv('/home/hobs/code/tangibleai/nlpia2/src/nlpia2/ch08/cnn-classify-text/data/trec.csv')
df.columns
df = df[reversed(df.columns)]
df



import pandas as pd
import numpy as np
t = np.arange(10000)
x = 10*t**2 + v0*t + sin(t/20/np.pi)
x = 10*t**2 + v0*t + sin(t/20/np.pi)
v0 = -3
x = 10*t**2 + v0*t + sin(t/20/np.pi)
x = 10*t**2 + v0*t + np.sin(t/20/np.pi)
x
df = pd.DataFrame()
df['t'] = t
df['y'] = x
df.plot(x='t', y='y')
import seaborn
from matplotlib import pyplot as plt
plt.show()
y = a*t**2 + v0*t + A*np.sin(t/20/np.pi)
y = a*t**2 + v0*t + A*np.sin(t/20/np.pi)
a = -9.8
y = a*t**2 + v0*t + A*np.sin(t/20/np.pi)
A = 1e3
y = a*t**2 + v0*t + A*np.sin(t/20/np.pi)
df['y'] = y
df.plot(x='t', y='y')
plt.show()
y = a*t**2 + v0*t + A*np.sin(t/20/np.pi)*t**2
a = 10
y = a*t**2 + v0*t + A*np.sin(t/20/np.pi)*t**2
df['y'] = y
df.plot(x='t', y='y') ; plt.show()
df.shift?
df['y'].shift(1)
df['y_t1'] = df['y'].shift(1)
X = df[['y']]
y = df['y_t1']
from sklearn.linear_model import Lasso
model = Lasso()
model.fit(X, y)
model.fit(X[1:], y[1:])
X = df[['y_t1']]
y = df['y;]
y = df['y']
model.fit(X[1:], y[1:])
model.score(X, y)
X = df[['y_t1']][1:]
y = df['y'][1:]
model.fit(X, y, num_iter=1000)
model.fit(X, y, n_iter=1000)
model.fit(X, y, niter=1000)
model = Lasso(n_iter=1000)
model = Lasso(niter=1000)
model = Lasso?
model = Lasso(max_iter=1000)
model.fit(X, y)
model = Lasso(max_iter=100000)
model.fit(X, y)
from sklearn.linear_model import LinearRegression
model = LinearRegression()
model.fit(X, y)
model.score(X, y)
hist -o -p -f ch07_rnn_synthetic_data.hist.ipy.md
hist -f ch07_rnn_synthetic_data.py



"""
If you add unimportant (frequent) words,
 the tf-idf vector search will still find the right text:

>>> question = 'What is overfitting a model?'
>>> question_vector = vectorizer.transform([question]).todense()
>>> np.array(question_vector).reshape(1,-1).dot(df_vecs.values.T).argmax()
51
>>> df['question'][51]
'What is overfitting?'

If you change unimportant (frequent) words,
 the tf-idf vector search will still find the right text:

>>> question = 'Who is overfitting a model?'
>>> question_vector = vectorizer.transform([question]).todense()
>>> question_vector.dot(df_vecs.values.T).argmax()
51

If you delete unimportant (frequent) words,
 the tf-idf vector search will still find the right text:

>>> question = 'overfitting'
>>> question_vector = vectorizer.transform([question]).todense()
>>> question_vector.dot(df_vecs.values.T).argmax()
51

>>> question = 'What is overfitting a model?'
>>> question_vector = vectorizer.transform([question]).todense()
>>> question_vector.dot(df_vecs.values.T).argmax()
51

If you change an important (infrequent) word, such as "overfit"
 it will find other questions that use that form of the word:

>>> question = 'What causes a model to overfit?'
>>> question_vector = vectorizer.transform([question]).todense()
>>> question_vector.dot(df_vecs.values.T).argmax()
35
>>> df['question'][35]
'Is a `LogisticRegression` more likely or less likely to overfit '
' to your data when compared to a `DecisionTree`?'

What was in this question_vector to cause it to match this FAQ question?

>>> question_vector[question_vector > 0]
matrix([[0.49665284, 0.73814899, 0.36844366, 0.26966885]])
>>> question_vector = pd.Series(np.array(question_vector)[0],
...     index=vectorizer.get_feature_names())
>>> question_vector[question_vector > 0]
model      0.496653
overfit    0.738149
to         0.368444
what       0.269669

>>> %run ch03_bot_sparse.py "how are you doing?"
Your question:
  how are you doing?
Most similar FAQ:
  What is a good package for doing topic analysis or natural language analytics?
Answer to that FAQ:
  The ldaviz package has good visualizations in addition to the Latent Dirichlet Allocation topic vector analysis.
"""
import sys
import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer

REPO_URL = 'https://gitlab.com/tangibleai/qary/-/raw/master'
FAQ_DIR = 'src/qary/data/faq'
FAQ_FILENAME = 'short-faqs.csv'
DS_FAQ_URL = '/'.join([REPO_URL, FAQ_DIR, FAQ_FILENAME])


class FAQ():
    def __init__(self, df, **kwargs):
        self.df = df
        self.vectorizer = TfidfVectorizer(**kwargs)
        self.vectorizer.fit(df['question'])
        self.q_vectors = self.vectorizer.transform(df['question'])
        # self.q_vectors = pd.DataFrame.sparse.from_spmatrix(
        #    self.q_vectors,
        #    columns=vectorizer.get_feature_names())
        # self.df = pd.concat([df, self.q_vectors], axis=1)

    def find_question_index(self, question):
        q_vector = self.vectorizer.transform([question])
        q_similarities = q_vector.dot(self.q_vectors.T)
        return q_similarities.argmax()

    def reply(self, question):
        idx = self.find_question_index(question)
        self.df['question'][idx]
        response = (
            f"Your question:\n  {question}\n"
            f"Most similar FAQ:\n  {df['question'][idx]}\n"
            f"Answer to that FAQ:\n  {df['answer'][idx]}\n"
        )
        return response


if __name__ == '__main__':
    if len(sys.argv) > 1:
        q = ' '.join(sys.argv[1:])
    else:
        q = "overfitting?"
    df = pd.read_csv(DS_FAQ_URL)
    faq = FAQ(df)
    idx = faq.find_question_index(q)
    # 51
    print(faq.reply(q))



""" Load custom word vectors into gensim language model (`nlp`)

### References
- https://stackoverflow.com/a/50091148/623735
"""

import spacy

# Path to google news vectors

google_news_path = "GoogleNews-vectors-negative300.bin.gz"

# Init blank english spacy nlp object
nlp = spacy.blank('en')

# Loop through range of all indexes, get words associated with each index.
# The words in the keys list will correspond to the order of the google embed matrix
keys = []
for idx in range(3000000):
    keys.append(model.index2word[idx])

# Set the vectors for our nlp object to the google news vectors
nlp.vocab.vectors = spacy.vocab.Vectors(data=model.syn0, keys=keys)

"""
>>> nlp.vocab.vectors.shape
(3000000, 300)
"""



from pathlib import Path
import re
import sys

HEADING_SYMBOL_LEVELS = {1: 0, 2: 1, 3: 2, 4: 3, 5: 4}
MAX_HEADING_SYMBOLS = max(HEADING_SYMBOL_LEVELS)
MIN_HEADING_SYMBOLS = min(HEADING_SYMBOL_LEVELS)

RE_HEADING = r'^([=]+\s*|:Chapter:\s?[0-9]+)(.+)$'

RE_HEADING = r'^([=]{'
RE_HEADING += str(MIN_HEADING_SYMBOLS) + ',' + str(MAX_HEADING_SYMBOLS)
RE_HEADING += r'}\s?|:Chapter:\s?[0-9]+)(.+)$'
CRE_HEADING = re.compile(RE_HEADING)


SYM_CHAPTER_TITLE = '= '  # "= Reasoning with word embeddings"
SYM_CHAPTER_NUM = ':Chapter:'  # ":Chapter: 6"


def make_path(p):
    return Path(p).expand_user().resolve().absolute()


def open_path(filepath, *args, **kwargs):
    return make_path(filepath).open(*args, **kwargs)


def extract_adoc_outline(filepath):
    outline_lines = []

    with open_path(filepath, 'rt') as fin:
        for line in fin:
            groups = CRE_HEADING.match(line)
            if groups and groups[1] and groups[2]:
                if groups[1].startswith('='):
                    level = len(groups[1].strip()) - 1
                    text = groups[2]
                    newlines = '\n' * int(level == 0)
                    outline_lines.append('  ' * level + newlines + "* " + text)
                elif groups[2].lower().startswith(SYM_CHAPTER_TITLE.lower()):
                    chapter_num = groups[2]
                    outline_lines.append('* ' + groups[1] + chapter_num)

    return outline_lines


def generate_book_outline(adoc_path, glob='Chapter*.adoc'):
    adoc_path = make_path(adoc_path)
    if adoc_path.is_file():
        adoc_paths = [adoc_path]
    else:
        adoc_paths = adoc_path.glob(glob)
    for p in adoc_paths:
        yield {str(p.with_suffix('').name): extract_adoc_outline(p)}


"""
$ make_outline.sh

```python
python -c '\
  with open("docs/headings.adoc") as fin: \
    lines = fin.readlines()
  ch_lines = []
  for i, line in enumerate(lines):
    if line.lower().startswith(':chapter:'):
      toks = line[-1].split()
      ch_lines[-1] = ' '.join(toks[:1] + [line.strip()] + toks[1:])
    else:
      ch_lines.append(line)
  with open("docs/headings.adoc", 'w') as fout: \
    fout.writeines(lines)
  '
```

```bash
newline =$'\n'


sed 's/=/#/g' docs / headings.adoc > docs / headings.md
sed 's/==== /      * /g' docs / headings.adoc > docs / outline.md
sed - i 's/=== /    * /g' docs / outline.md
sed - i 's/== /  * /g' docs / outline.md
sed - i 's/= /'"\\${newline}"'* /g' docs / outline.md

# rm -f headings.adoc

# pandoc --atx-headers     --verbose     --wrap=none     --toc     --reference-links     -s -o outline.adoc     outline.md
"""


if __name__ == '__main__':
    adoc_path = 'manuscript/adoc'
    glob = 'Chapter*.adoc'
    if len(sys.argv) > 1:
        adoc_path = sys.argv[1]
    if len(sys.argv) > 2:
        glob = sys.argv[2]
    chapter_outlines = list(generate_book_outline(adoc_path, glob=glob))



"""Spelling Corrector in Python 3; see http://norvig.com/spell-correct.html

Copyright (c) 2007-2016 Peter Norvig
MIT license: www.opensource.org/licenses/mit-license.php

References:
- https://norvig.com/spell-correct.html
- http://static.googleusercontent.com/external_content/untrusted_dlcp/research.google.com/en/us/pubs/archive/36180.pdf
- http://citeseerx.ist.psu.edu/viewdoc/download;jsessionid=52A3B869596656C9DA285DCE83A0339F?doi=10.1.1.146.4390&rep=rep1&type=pdf
"""
# ############### Spelling Corrector
from collections import Counter
import re
import requests


# any document of space-delimitted words will do
BIG_TXT = requests.get('https://norvig.com/big.txt').text
WORDS = re.findall(r'\w+', BIG_TXT.lower())
WORDS = Counter(WORDS)


def P(word, N=sum(WORDS.values())):
    "Probability of `word`."
    return WORDS[word] / N


def correction(word):
    "Most probable spelling correction for word."
    return max(candidates(word), key=P)


def candidates(word):
    "Generate possible spelling corrections for word."
    return (known([word]) or known(edits1(word)) or known(edits2(word)) or [word])


def known(words):
    "The subset of `words` that appear in the dictionary of WORDS."
    return set(w for w in words if w in WORDS)


def edits1(word):
    "All edits that are one edit away from `word`."
    letters = 'abcdefghijklmnopqrstuvwxyz'
    splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
    deletes = [L + R[1:] for L, R in splits if R]
    transposes = [L + R[1] + R[0] + R[2:] for L, R in splits if len(R) > 1]
    replaces = [L + c + R[1:] for L, R in splits if R for c in letters]
    inserts = [L + c + R for L, R in splits for c in letters]
    return set(deletes + transposes + replaces + inserts)


def edits2(word):
    "All edits that are two edits away from `word`."
    return (e2 for e1 in edits1(word) for e2 in edits1(e1))

# Test Code


def unit_tests():
    assert correction('speling') == 'spelling'              # insert
    assert correction('korrectud') == 'corrected'           # replace 2
    assert correction('bycycle') == 'bicycle'               # replace
    assert correction('inconvient') == 'inconvenient'       # insert 2
    assert correction('arrainged') == 'arranged'            # delete
    assert correction('peotry') == 'poetry'                  # transpose
    assert correction('peotryy') == 'poetry'                 # transpose + delete
    assert correction('word') == 'word'                     # known
    assert correction('quintessential') == 'quintessential'  # unknown
    # assert words('This is a TEST.') == ['this', 'is', 'a', 'test']
    # assert Counter(words('This is a test. 123; A TEST this is.')) == (
    #     Counter({'123': 1, 'a': 2, 'is': 2, 'test': 2, 'this': 2}))
    assert len(WORDS) == 32192
    assert sum(WORDS.values()) == 1115504
    assert WORDS.most_common(10) == [
        ('the', 79808),
        ('of', 40024),
        ('and', 38311),
        ('to', 28765),
        ('in', 22020),
        ('a', 21124),
        ('that', 12512),
        ('he', 12401),
        ('was', 11410),
        ('it', 10681)]
    assert WORDS['the'] == 79808
    assert P('quintessential') == 0
    assert 0.07 < P('the') < 0.08
    return 'unit_tests pass'


def spelltest(tests, verbose=False):
    "Run correction(wrong) on all (right, wrong) pairs; report results."
    import time
    start = time.clock()
    good, unknown = 0, 0
    n = len(tests)
    for right, wrong in tests:
        w = correction(wrong)
        good += (w == right)
        if w != right:
            unknown += (right not in WORDS)
            if verbose:
                print('correction({}) => {} ({}); expected {} ({})'
                      .format(wrong, w, WORDS[w], right, WORDS[right]))
    dt = time.clock() - start
    print('{:.0%} of {} correct ({:.0%} unknown) at {:.0f} words per second '
          .format(good / n, n, unknown / n, n / dt))


def Testset(lines):
    "Parse 'right: wrong1 wrong2' lines into [('right', 'wrong1'), ('right', 'wrong2')] pairs."
    return [(right, wrong)
            for (right, wrongs) in (line.split(':') for line in lines)
            for wrong in wrongs.split()]


if __name__ == '__main__':
    print(unit_tests())
    SPELL_TESTSET1 = requests.get('https://norvig.com/spell-testset1.txt').text
    SPELL_TESTSET2 = requests.get('https://norvig.com/spell-testset2.txt').text
    spelltest(Testset(SPELL_TESTSET1.split('\n')))
    spelltest(Testset(SPELL_TESTSET2.split('\n')))



>>> import spacy

>>> spacy_de = spacy.load('de')
>>> spacy_en = spacy.load('en')

>>> def tokenize_de(text):
...     return [tok.text for tok in spacy_de.tokenizer(text)]

>>> def tokenize_en(text):
...     return [tok.text for tok in spacy_en.tokenizer(text)]
from pytorch import nn
from torch import nn
>>> class PositionalEncoding(nn.Module):
...     def __init__(self, d_model, dropout=0.1, max_len=5000):
...         super(PositionalEncoding, self).__init__()
...         self.dropout = nn.Dropout(p=dropout)

...         pe = torch.zeros(max_len, d_model)
...         position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
...         div_term = torch.exp(torch.arange(0, d_model, 2).float() *
...                              (-math.log(10000.0) / d_model))
...         pe[:, 0::2] = torch.sin(position * div_term)
...         pe[:, 1::2] = torch.cos(position * div_term)
...         pe = pe.unsqueeze(0).transpose(0, 1)
...         self.register_buffer('pe', pe)

...     def forward(self, x):
...         x = x + self.pe[:x.size(0), :]
...         return self.dropout(x)
from spacy import cli
cli.download("de")
>>> import spacy

>>> spacy_de = spacy.load('de')
>>> spacy_en = spacy.load('en')

>>> def tokenize_de(text):
...     return [tok.text for tok in spacy_de.tokenizer(text)]

>>> def tokenize_en(text):
...     return [tok.text for tok in spacy_en.tokenizer(text)]
spacy.load("de")
nlp_de = spacy.load('de_core_news_sm')
nlp_de("Sprechen Sie Deutche")
doc = nlp_de("Sprechen Sie Deutche")
for tok in doc:
    print(tok.text)
for tok in doc:
    print(tok.pos_, tok.text)
for tok in doc:
    print(tok.vector[:3], tok.pos_, tok.text)
>>> import spacy

>>> spacy_de = spacy.load('de_core_news_sm')
>>> spacy_en = spacy.load('en')

>>> def tokenize_de(text):
...     return [tok.text for tok in spacy_de.tokenizer(text)]

>>> def tokenize_en(text):
...     return [tok.text for tok in spacy_en.tokenizer(text)]
>>> import spacy

>>> spacy_de = spacy.load('de_core_news_sm')
>>> spacy_en = spacy.load("en_core_web_sm")

>>> def tokenize_de(text):
...     return [tok.text for tok in spacy_de.tokenizer(text)]

>>> def tokenize_en(text):
...     return [tok.text for tok in spacy_en.tokenizer(text)]
>>> import spacy

>>> spacy_de = spacy.load('de_core_news_sm')
>>> spacy_en = spacy.load("en_core_web_md")

>>> def tokenize_de(text):
...     return [tok.text for tok in spacy_de.tokenizer(text)]

>>> def tokenize_en(text):
...     return [tok.text for tok in spacy_en.tokenizer(text)]
cli.download("en_core_web_sm")
>>> import spacy

>>> spacy_de = spacy.load('de_core_news_sm')
>>> spacy_en = spacy.load("en_core_web_md")

>>> def tokenize_de(text):
...     return [tok.text for tok in spacy_de.tokenizer(text)]

>>> def tokenize_en(text):
...     return [tok.text for tok in spacy_en.tokenizer(text)]
>>> import spacy

>>> spacy_de = spacy.load('de_core_news_sm')
>>> spacy_en = spacy.load("en_core_web_sm")

>>> def tokenize_de(text):
...     return [tok.text for tok in spacy_de.tokenizer(text)]

>>> def tokenize_en(text):
...     return [tok.text for tok in spacy_en.tokenizer(text)]
>>> import torchtext
>>> from torchtext.datasets import Multi30k
>>> from torchtext.data import Field, BucketIterator

>>> SRC = Field(tokenize = tokenize_de,
...             init_token = '<sos>',  
...             eos_token = '<eos>',
...             lower = True,
...             batch_first = True)

>>> TRG = Field(tokenize = tokenize_en,
...             init_token = '<sos>',
...             eos_token = '<eos>',
...             lower = True,
...             batch_first = True)
import torchtext.data
dir(torchtext.data)
dir(torchtext.data.functional)
import torchtext.data.BucketIterator
from torchtext.data import *
who
from torchtext.data import field
import torchtext
torchtext.__version__
from torchtext.datasets import field
from torchtext.datasets import Field
from torchtext.datasets import IWSLT2017
from torchtext.datasets import Multi30k
>> from torch import Tensor
>>> from typing import Optional, Any

>>> class CustomDecoderLayer(nn.TransformerDecoderLayer):
...     def forward(self, tgt: Tensor, memory: Tensor, tgt_mask: Optional[Tensor] = None,
...                 memory_mask: Optional[Tensor] = None,
...                 tgt_key_padding_mask: Optional[Tensor] = None,
...                 mem_key_padding_mask: Optional[Tensor] = None) -> Tensor:
...         """Same as DecoderLayer but returns multi-head attention weights.
...         """
...         tgt2 = self.self_attn(tgt, tgt, tgt, attn_mask=tgt_mask,
...                               key_padding_mask=tgt_key_padding_mask)[0]
...         tgt = tgt + self.dropout1(tgt2)
...         tgt = self.norm1(tgt)
...         tgt2, attention_weights = self.multihead_attn(tgt, memory, memory,  
...                                                       attn_mask=memory_mask,
...                                                       key_padding_mask=mem_key_padding_mask,
...                                                       need_weights=True)
...         tgt = tgt + self.dropout2(tgt2)
...         tgt = self.norm2(tgt)
...         tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
...         tgt = tgt + self.dropout3(tgt2)
...         tgt = self.norm3(tgt)
...         return tgt, attention_weights
>>> from torch import Tensor
>>> from typing import Optional, Any

>>> class CustomDecoderLayer(nn.TransformerDecoderLayer):
...     def forward(self, tgt: Tensor, memory: Tensor, tgt_mask: Optional[Tensor] = None,
...                 memory_mask: Optional[Tensor] = None,
...                 tgt_key_padding_mask: Optional[Tensor] = None,
...                 mem_key_padding_mask: Optional[Tensor] = None) -> Tensor:
...         """Same as DecoderLayer but returns multi-head attention weights.
...         """
...         tgt2 = self.self_attn(tgt, tgt, tgt, attn_mask=tgt_mask,
...                               key_padding_mask=tgt_key_padding_mask)[0]
...         tgt = tgt + self.dropout1(tgt2)
...         tgt = self.norm1(tgt)
...         tgt2, attention_weights = self.multihead_attn(tgt, memory, memory,  
...                                                       attn_mask=memory_mask,
...                                                       key_padding_mask=mem_key_padding_mask,
...                                                       need_weights=True)
...         tgt = tgt + self.dropout2(tgt2)
...         tgt = self.norm2(tgt)
...         tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
...         tgt = tgt + self.dropout3(tgt2)
...         tgt = self.norm3(tgt)
...         return tgt, attention_weights
decoder = CustomDecoderLayer()
decoder = CustomDecoderLayer(nhead=3)
decoder = CustomDecoderLayer?
decoder = CustomDecoderLayer(50_000, 3)
decoder = CustomDecoderLayer(300, 100)
import torch
from torch import Tensor
import nump as np
import numpy as np
np.random.randn()
np.random.randn((300, 100))
np.random.randn(300, 100)
Tensor(np.random.randn(300, 100))
inp = Tensor(np.random.randn(300, 100))
decoder.forward(inp)
history
%hist -o -p -f ch12.hist.ipy
%hist -f ch12.hist.py



import argparse
import os
import copy
from graphviz import Digraph, Graph
import yaml
import sys
from pathlib import Path
import shutil
import logging
from nlpia2.constants import DATA_DIR

__version__ = None
try:
    from nlpia2 import __version__  # noqa
except ImportError:
    pass
try:
    from nlpia2.constants import __version__  # noqa
except ImportError:
    pass


log = logging.getLogger(__name__)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=loglevel, stream=sys.stdout,
                        format=logformat, datefmt="%Y-%m-%d %H:%M:%S")


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """


CLASSES = dict(digraph=Digraph, graph=Graph)

# __file__ defined so script can be %run within ipython
SCRIPT_WORKING_DIR = os.getcwd()
try:
    __file__ = __file__
except NameError:
    __file__ = SCRIPT_WORKING_DIR / f'{__name__}.py'


def find_code_dir(filepath=__file__):
    CODE_DIR = Path(filepath).resolve().absolute()
    for i in range(4):
        if not CODE_DIR or not CODE_DIR.name or CODE_DIR.name in ['code', 'nlpia2']:
            break
        # print(f"NOT code dir: {CODE_DIR}")
        CODE_DIR = CODE_DIR.parent
    return CODE_DIR


CODE_DIR = find_code_dir()


def find_repo_dir(dirpath=CODE_DIR):
    REPO_DIR = dirpath.parent
    for i in range(4):
        if REPO_DIR.name in ['nlpia-manuscript', 'nlpia2']:
            break
        # print(f"not repo dir: {REPO_DIR}")
        REPO_DIR = REPO_DIR.parent
    return REPO_DIR


REPO_DIR = find_repo_dir()


def find_dest_dir(home_code_dir=REPO_DIR.parent.parent):
    print('find_dest')
    print(f'home_code_dir: {home_code_dir}')
    assert home_code_dir.name == 'code'
    MANUSCRIPT_DIR = home_code_dir / 'tangibleai' / 'nlpia-manuscript' / 'manuscript'
    assert MANUSCRIPT_DIR.is_dir()
    dest_dir = MANUSCRIPT_DIR / 'images'
    assert dest_dir.is_dir()
    print(f'dest_dir: {dest_dir}')
    return dest_dir


DEST_DIR = find_dest_dir()


def wrap_text(text, max_line_width=10):
    text = str(text)
    lines = []
    words = text.split()
    if len(text) < max_line_width:
        return text
    for i, w in enumerate(words):
        if len(w) < 3:
            if i:
                lines[-1] += f' {w}'
            else:
                if len(words) > 1:
                    words[i + 1] = f'{w} ' + words[i + 1]
                else:
                    lines.append(w)
        else:
            lines.append(w)
    return '\n'.join(lines)


def load_graphviz(filepath=None, engine='sfdp', attr=None, node_attr=None):
    """ Load yaml file into dictionary of graphviz args for pygraphviz

    Inputs:
      engine (https://graphviz.org/docs/layouts/):
        dot: hierarchies that look like org charts (best for Digraphs)
        neato: minimum energy layoud (Brownian motion? Lagrange functions?)
        circo: all nodes on edge of circle
        fdp: force-directed graph that reduces spring force instead of energy (like neato), order of "appearance" out from center (first nodes)
        osage: hierarchical clusters of subgraphs
        patchwork: squarified treemap layout
        sfdp: multiscale fdp for large graphs with many nodes

    """
    filepath = filepath or CODE_DIR / 'data' / 'nlp-applications-graphviz.yml'
    engine = engine or 'sfdp'  # neato, fdp, sfdp, dot
    # u = Digraph('unix', filename='unix.gv',
    #             node_attr={'color': 'lightblue2', 'style': 'filled'})
    ATTR = dict(
        # engine='fdp',
        rankdir='LR',
        # layout="neato",
        # size='6,6',
        #     nodesep=1,
        #     ranksep=1,
    )
    attr = attr or ATTR

    node_attr = node_attr or dict(shape='plaintext')

    filepath = str(filepath)
    with open(filepath) as fin:
        y = yaml.full_load(fin)
    attr = copy.deepcopy(dict(attr))
    node_attr = copy.deepcopy(node_attr)
    log.warning(f'yaml filepath: {filepath}')
    Klass = CLASSES[y.get('class', 'digraph').lower()]
    name = ''.join(str(filepath).split('.')[:-1])
    name = str(y.get('name') or name)
    engine = str(y.get('engine') or engine)
    log.warning(f'engine: {engine}')
    attr.update(y.get('attr', {}))
    node_attr.update(y.get('node_attr', {}))
    g = Klass(name, filename=name + '.gv',
              engine=engine, node_attr=node_attr)
    g.attr(**attr)
    # print(g)
    for i, node in enumerate(y.get('nodes', [])):
        label, kwargs = None, {}
        if isinstance(node, str):
            label = node
        elif len(node) == 1:
            label = str(node[0])
        elif len(node) == 2:
            if isinstance(node[1], dict):
                label = str(node[0])
                kwargs = dict(node[1])
            else:
                log.warning(f"Unable to parse node #{i}: {node}")
        if label is not None:
            # if label is just whitespace, then don't give it a box/oval/circle shape
            if not label.strip():
                kwargs.update({'shape': 'plaintext'})
            log.warning(f"{label}, {kwargs}")
            if len(kwargs):
                g.node(wrap_text(label), **kwargs)
            else:
                g.node(wrap_text(label))

    unique_edges = set()
    for e in y.get('edges', []):
        edge_str = str(e)
        if edge_str in unique_edges:
            continue
        unique_edges.add(edge_str)
        # print(e)
        if len(e) == 2:
            g.edge(wrap_text(e[0]), wrap_text(e[1]))
        elif len(e) == 3:
            if isinstance(e[2], str):
                e[2] = dict(label=wrap_text(e[2]))
            g.edge(wrap_text(e[0]), wrap_text(e[1]), **e[2])
    return g


def brittle_parse_args(args=None):
    """ Assumes 3 cli args are YAML_FILE CHNUM DEST_DIR  """
    args = args or sys.argv[1:]
    dest_dir = DEST_DIR if len(args) < 3 else Path(args[2])
    assert dest_dir.is_dir()
    chnum = '' if len(args) < 2 else str(args[1])
    dest_dir /= str(chnum)
    assert dest_dir.is_dir()
    yaml_filepath = None if len(args) < 1 else Path(args[0]).expanduser().resolve().absolute()
    print(f'yaml_filepath: {yaml_filepath}')
    return dict(dest_dir=dest_dir, yaml_filepath=yaml_filepath, loglevel=logging.WARNING)


def parse_args(args=None):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      argparse.Namespace: command line parameters as attributes of namespace object
    """
    parser = argparse.ArgumentParser(
        description="Transpile yaml->graphviz.dot then render svg/png files")
    parser.add_argument(
        "--version",
        action="version",
        version="yaml_graphviz {ver}".format(ver=__version__))
    parser.add_argument(
        dest="yaml_file",
        nargs='?',
        help="Path to yaml file containing graph specification for graphviz diagram",
        type=Path,
        metavar="YAML_FILE")
    parser.add_argument(
        dest="dest_subdir",
        nargs='?',
        help="Subdirectory within dest_dir (e.g. `ch05`)",
        type=Path,
        metavar="CHXX")
    parser.add_argument(
        '-d',
        '--dest',
        dest="dest_dir",
        help="images directory to render svg and png to",
        default=Path(DEST_DIR),
        type=Path,
        metavar="DEST_DIR")
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO)
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG)
    args = sys.argv[1:] if args is None else args
    args = parser.parse_args(args)
    return dict(yaml_filepath=args.yaml_file,
                dest_dir=args.dest_dir,
                loglevel=args.loglevel)


def render_yaml_graphviz(yaml_filepath=None, dest_dir=DEST_DIR):
    if yaml_filepath is None:
        yaml_filepath = next(DATA_DIR.glob('*graphviz.yml'))
    assert yaml_filepath.is_file()
    g = load_graphviz(filepath=yaml_filepath)
    name = str(g.name)
    for ext in ['svg', 'png']:
        g.render(filename=name, cleanup=1, view=0, format=ext)
        # g.save()
        # g.view()

        dest = DEST_DIR / (name + '.' + ext)
        log.warning(f'svg filepath: {dest}')
        try:
            dest.resolve().absolute().unlink()
        except FileNotFoundError:
            pass
        shutil.move(name + '.' + ext, str(dest.resolve().absolute()))
    # g.view()


if __name__ == '__main__':
    # FIXME: kwargs = parse_args()
    kwargs = parse_args()
    setup_logging(kwargs.pop('loglevel'))
    print(kwargs)
    render_yaml_graphviz(**kwargs)



import pandas as pd
from matplotlib import pyplot as plt
import seaborn
from pathlib import Path

from nlpia2.init import DATA_DIR, IMAGES_DIR


FILENAME_ROOT = Path(__file__).with_suffix('').name
CH_STR = 'ch07'
SHOW = False


def render(
        filename_root=FILENAME_ROOT, show=SHOW):
    seaborn.set_theme('notebook')
    seaborn.set_style('whitegrid')

    df = pd.read_csv(DATA_DIR / filename_root, index_col=0)
    figsize = (10, 6)
    fig = plt.figure(figsize=figsize)
    df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
    df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
    df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
    plt.grid('on')
    plt.xlabel('Time (sec)')
    plt.ylabel('Signal amplitude (-1 to +1)')
    plt.legend()
    fig.savefig(IMAGES_DIR / CH_STR / filename_root + '.png')
    if show:
        plt.show(block=False)
    return fig


if __name



""" Sense2vec (2015 paper) creates better embeddings, especially for noun-phrases

### References
- https://github.com/explosion/sense2vec
- [sense2vec - A Fast and Accurate Method for Word Sense Disambiguation In Neural Word Embeddings](
https://arxiv.org/abs/1511.06388 by Andrew Trask, Phil Michalak, John Liu)
"""

from sense2vec import Sense2Vec

s2v = Sense2Vec().from_disk("/path/to/s2v_reddit_2015_md")
query = "natural_language_processing|NOUN"
assert query in s2v
vector = s2v[query]
freq = s2v.get_freq(query)
most_similar = s2v.most_similar(query, n=3)
# [('machine_learning|NOUN', 0.8986967),
#  ('computer_vision|NOUN', 0.8636297),
#  ('deep_learning|NOUN', 0.8573361)]



from ch05.fake_news_logistic_regression_title_len import *
# import numpy as np

try:
    assert len(__file__)
except (AssertionError, NameError):
    __file__ = "fake_news_histogram_title_len_red_blue.py"

isfake = y > 0
sns.histplot(X['title_len'][isfake], color='r', bins=list(range(0, 300, 5)))
sns.histplot(X['title_len'][~isfake], color='b', bins=list(range(0, 300, 5)))
image_filepath = CH05_IMAGES_DIR / (Path(__file__).name[:-2] + 'png')
print(f'Saving logistic regression plot to {image_filepath}')
plt.savefig(image_filepath)
plt.show()



import spacy
from spacy import displacy
from pathlib import Path
from itertools import chain

import os
import subprocess
# import json
import logging

log = logging.getLogger(__name__)


CODE_DIR = Path(__file__).resolve().absolute()
for i in range(4):
    if CODE_DIR.name in ['code', 'nlpia2']:
        break
    # print(f"NOT code dir: {CODE_DIR}")
    CODE_DIR = CODE_DIR.parent

REPO_DIR = CODE_DIR.parent
for i in range(4):
    if REPO_DIR.name in ['nlpia-manuscript', 'nlpia2']:
        break
    # print(f"not repo dir: {REPO_DIR}")
    REPO_DIR = REPO_DIR.parent

HOME_CODE_DIR = REPO_DIR.parent.parent
print(HOME_CODE_DIR)
assert HOME_CODE_DIR.name == 'code'
MANUSCRIPT_DIR = HOME_CODE_DIR / 'tangibleai' / 'nlpia-manuscript' / 'manuscript'
assert MANUSCRIPT_DIR.is_dir()
IMAGE_DIR = MANUSCRIPT_DIR / 'images'
assert IMAGE_DIR.is_dir()
SCRIPT_WORKING_DIR = os.getcwd()


def run(command, chdir=None):
    if chdir:
        log.info(f'Temporarily changing working directory to {chdir}')
        initial_cwd = os.getcwd()
        os.chdir(chdir)

    log.warning(f'Running: {" ".join(command)}')
    output = subprocess.run(command, capture_output=True)

    if chdir:
        os.chdir(initial_cwd)

    return {
        'stderr': output.stderr.decode("utf-8").splitlines(),
        'stdout': output.stderr.decode("utf-8").splitlines()
    }


def render_svg(filepath):
    # exit_code = subprocess.call(cmd, shell=True)  # exit_code == 0 if successful
    filepath_noext = '.'.join(str(filepath).split('.')[:-1])
    # deprecated: cmd = f'inkscape --without-gui {filepath_noext}.svg -o {filepath_noext}.png'.split()
    cmd = f'inkscape {filepath_noext}.svg -o {filepath_noext}.png'.split()
    return run(command=cmd, chdir=None)


if __name__ == '__main__':
    # print(Path(__file__).resolve().absolute())
    print()
    print('-' * 70)
    print(Path(__file__).name)
    text = ("Trust me, though, the words were on their way, and when "
            "they arrived, Liesel would hold them in her hands like "
            "the clouds, and she would wring them out, like the rain. ")
    texts = [text]
    texts.append("There's no such thing as survival of the fittest. "
                 "Survival of the most adequate, maybe. ")

    try:
        nlp = spacy.load('en_core_web_md')
    except Exception:
        spacy.cli.download('en_core_web_md')
        nlp = spacy.load('en_core_web_md')

    docs = [nlp(txt) for txt in texts]
    sentences = list(chain([list(d.sents) for d in docs]))
    print(sentences)
    options = {
        'font': 'Arial 24',
        # 'fine_grained': True,
        'add_lemma': True,
        # 'compact': True,
        'arrow_stroke': 3,  # 2
        'word_spacing': 35,  # 45 (between arrow and word)
        'distance': 150,        # 175 (between words)
    }

    for slist in sentences:
        for s in slist:
            basename = '-'.join([w.text.lower() for w in s[:5] if len(w.text) > 2])

            html = displacy.render(s, style="dep", page=True, options=options)
            filename = basename + '.html'
            filepath = IMAGE_DIR / 'ch02' / filename
            with open(filepath, 'w') as f:
                f.write(html)

            svg = displacy.render(s, style="dep", options=options)
            filename = basename + '.svg'
            filepath = IMAGE_DIR / 'ch02' / filename
            with open(filepath, 'w') as f:
                f.write(svg)

            render_svg(filepath)

    svg = displacy.render(sentences[-1][-1], style="dep", options=options)
    filename = 'survival-of-adequate-sentence-diagram.svg'
    with open(IMAGE_DIR / 'ch02' / filename, 'w') as f:
        f.write(svg)
    print('-' * 70)
    print()


r"""
>>> text = ("Trust me, though, the words were on their way, and when "
...         "they arrived, Liesel would hold them in her hands like "
...         "the clouds, and she would wring them out, like the rain.")
>>> tokens = text.split()
>>> tokens[:8]
['Trust', 'me,', 'though,', 'the', 'words', 'were', 'on', 'their']
>>> import re
>>> pattern = r'\w+(?:\'\w+)?|[^\w\s]'  # <1>
>>> texts = [text]
>>> texts.append("There's no such thing as survival of the fittest."
...              "Survival of the most adequate, maybe.")
>>> tokens = list(re.findall(pattern, texts[-1]))
>>> tokens[:8]
["There's", 'no', 'such', 'thing', 'as', 'survival', 'of', 'the']
>>> tokens[8:16]
['fittest', '.', 'Survival', 'of', 'the', 'most', 'adequate', ',']
>>> tokens[16:]
['maybe', '.']
>>> import numpy as np  # <1>
>>> vocab = sorted(set(tokens))  # <2>
>>> ' '.join(vocab[:12])  # <3>
", . Survival There's adequate as fittest maybe most no of such"
>>> num_tokens = len(tokens)
>>> num_tokens
18
>>> vocab_size = len(vocab)
>>> vocab_size
15
texts
>>> import re
>>> pattern = r'\w+(?:\'\w+)?|[^\w\s]'  # <1>
>>> texts = [text]
>>> texts.append("There's no such thing as survival of the fittest. "
...              "Survival of the most adequate, maybe.")
"""



import pandas as pd
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer

pd.read_html('https://www.ssa.gov/oact/babynames/decades/names2010s.html')
for year in range(1940, 2020, 10):
    print(year)
for year in range(1940, 2020, 10):
    print(year)
    url = f'https://www.ssa.gov/oact/babynames/decades/names{year}s.html'
    print(url)
dfs = []
for year in range(1940, 2020, 10):
    print(year)
    url = f'https://www.ssa.gov/oact/babynames/decades/names{year}s.html'
    print(url)
    dfs.extend(pd.read_html(url))
len(dfs)
df = pd.concat(dfs)
len(df)
df.describe()
df.describe(include='all')
for year in range(1800, 2020, 10):
    print(year)
    url = f'https://www.ssa.gov/oact/babynames/decades/names{year}s.html'
    print(url)
    dfs.extend(pd.read_html(url))
for year in range(1800, 2020, 10):
    print(year)
    url = f'https://www.ssa.gov/oact/babynames/decades/names{year}s.html'
    print(url)
    try:
        dfs.extend(pd.read_html(url))
    except HTTPError:
        pass
for year in range(1800, 2020, 10):
    print(year)
    url = f'https://www.ssa.gov/oact/babynames/decades/names{year}s.html'
    print(url)
    try:
        dfs.extend(pd.read_html(url))
    except pd.HTTPError:
        pass
import lxml
from lxml import *
who
dir(lxml)
dir(lxml.html)
dir(lxml.html.defs)
dir(lxml.html)
dir(lxml.html.Classes)
dir(lxml.html.open_http_urllib)
dir(pd)
dir(pd.util)
dir(pd.util._exceptions)
for year in range(1800, 2020, 10):
    print(year)
    url = f'https://www.ssa.gov/oact/babynames/decades/names{year}s.html'
    print(url)
    try:
        dfs.extend(pd.read_html(url))
    except Exception:
        pass
len(dfs)
df2 = pd.concat(dfs)
df2.describe(include='all')
pd.options.display.max_columns = 1000
df2.describe(include='all')
df.head()
df.sample(20)
for year in range(1800, 2020, 10):
    print(year)
    url = f'https://www.ssa.gov/oact/babynames/decades/names{year}s.html'
    print(url)
    try:
        dfs.extend(pd.read_html(url))
        for df in dfs:
            df['decade'] = year
    except Exception:
        pass
    sleep(2)
!unzip / home / hobs / Downloads / names.zip
pd.read_csv('Names_2010Census.csv')
names = _
names.describe(include='all')
!unzip / home / hobs / Downloads / len - us - first - names - database.zip
pd.read_csv('len-us-first-names-database/original/Common_Surnames_Census_2000.csv ')
pd.read_csv('len-us-first-names-database/original/Common_Surnames_Census_2000.csv')
pd.read_csv('len-us-first-names-database/data/ssa_names_db.csv')
pd.read_csv('/home/hobs/Downloads/baby-names-state.csv')
df = _
df.to_csv('.nlpia2-data/baby-names-state.csv.gz', index=False)
df.describe(include='all')
pd.read_csv('/home/hobs/Downloads/baby-names-territories.csv')
dft = _
dft['state_abb'] = dft['territory_code']
del dft['territory_code']
dft.head()
df.head()
df = pd.concat([df, dft])
df.head()
df.describe(include='all')import pandas as pd
import numpy as np
from sklearn.feature_extraction.textimport pandas as pd
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer import TfidfVectorizer
dft.describe(include='all')
df['name'].apply(lambda s: ' ' in s).sum()
from sklearn.feature_extraction.text import TfidfVectorizer
vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
vectorizer.fit(df['name'])


# GOOD

df = pd.read_csv('.nlpia2-data/baby-names-region.csv.gz')
df['freq'] = 0.
for y in df['year'].unique():
    g = df[y == df['year']]
    tot = g['count'].sum()
    mask = y == df['year']
    df['freq'][mask] = df['count'][mask] / df['count'][mask].values.sum()
    print(y, tot, df['freq'][mask].sum())
df['freq'].sum().round(10)
df['year'].nunique()
df.to_csv('.nlpia2-data/baby-names-region.csv.gz', index=False)


# using preprocessed data
df.head()
df.columns = list(df.columns)[:-1] + ['frequency']
model = LogisticRegression(class_weight=df['frequency'])
vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
vectorizer.fit(df['name'])
vecs = vectorizer.transform(df['name'])
istrain = np.random.rand(len(vecs)) < .9

istrain = np.random.rand(len(vecs)) < .9
istrain = np.random.rand(vecs.shape[0]) < .9
X_train = vecs[istrain]
y_train = df['sex'][istrain]
model.fit(X_train, y_train)
istrain.shape
y_train.head()
model.fit(X_train.values, y_train.values)
type(X_train)
model.fit(vecs, df['sex'].values)
model.fit(vecs, (df['sex'].values == 'F'))
sex = df['sex'].values == 'F'
sex.sum() / len(sex)
model = LogisticRegression()
model.fit(vecs, df['sex'])
LogisticRegression?
LogisticRegression?
model.fit(vecs, df['sex'], sample_weight=df['frequency'])


# can start here
# model = LogisticRegression(class_weight=df10000['count'])
model.fit(vecs, df['sex'][istrain], sample_weight=df['frequency'][istrain])
model.fit(vecs[istrain], df['sex'][istrain], sample_weight=df['frequency'][istrain])
df.columns = list(df.columns)[:-1] + ['freq']


model.fit(vecs[istrain], df['sex'][istrain], sample_weight=df['freq'][istrain])
model.score(vecs[istrain], df['sex'][istrain], sample_weight=df['freq'][istrain])
model.score(vecs[~istrain], df['sex'][~istrain], sample_weight=df['freq']~[istrain])
model.score(vecs[~istrain], df['sex'][~istrain], sample_weight=df['freq'][~istrain])
model.predict(vectorizer.transform(['Hobson', 'Maria', 'Aditi', 'Vish', 'Jessica']))
names = ['Hobson', 'Maria', 'Aditi', 'Vish', 'Jessica', 'Mohammed', 'Olessya']
pd.Series(model.predict(vectorizer.transform(names)), index=names)
names = ['Maria', 'Aditi', 'Jessica', 'Olessya', 'Una', 'Hanna', 'Winnie', 'Olessya',
         'Sylvia', 'Vish', 'Mohammed', 'Jon', 'John', 'Ted', 'Kazuma', 'Meijke']
pd.Series(model.predict(vectorizer.transform(names)), index=names)
from torch import nn
history - o - p - f src / nlpia2 / ch05 / name_sex.md
history - f src / nlpia2 / ch05 / name_sex.py


# Just the working code
from tqdm import tqdm
import pandas as pd
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression

df = pd.read_csv('.nlpia2-data/baby-names-region.csv.gz')
istrain = np.random.rand(len(df)) < .9
vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
vectorizer.fit(tqdm(df['name'][istrain]))
vecs = vectorizer.transform(tqdm(df['name']))

model = LogisticRegression(max_iter=2000, class_weight='balanced')
model.fit(vecs[istrain], df['sex'][istrain], sample_weight=df['freq'][istrain])


names = ['Maria', 'Aditi', 'Jessica', 'Olessya', 'Una', 'Hanna', 'Winnie', 'Olessya',
         'Sylvia', 'Vish', 'Mohammed', 'Jon', 'John', 'Ted', 'Kazuma', 'Meijke', 'Kemal']
pd.Series(model.predict(vectorizer.transform(names)), index=names)


# Mob Programming Session
from tqdm import tqdm
import pandas as pd
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression

df = pd.read_csv('.nlpia2-data/baby-names-region.csv.gz')
istrain = np.random.rand(len(df)) < .9
vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
vectorizer.fit(tqdm(df['name'][istrain]))
vecs = vectorizer.transform(tqdm(df['name']))
model = LogisticRegression(max_iter=2000, class_weight='balanced')
model.fit(vecs[istrain], df['sex'][istrain], sample_weight=df['freq'][istrain])
pd.Series(model.predict(vectorizer.transform(names)), index=names)
names = ['Maria', 'Aditi', 'Jessica', 'Olessya', 'Una', 'Hanna', 'Winnie', 'Olessya',
         'Sylvia', 'Vish', 'Mohammed', 'Jon', 'John', 'Ted', 'Kazuma', 'Meijke', 'Kemal']
pd.Series(model.predict(vectorizer.transform(names)), index=names)
model = LogisticRegression(max_iter=10000)
model.fit(vecs[istrain], df['sex'][istrain], sample_weight=df['freq'][istrain])
pd.Series(model.predict(vectorizer.transform(names)), index=names)
(df['name'] == 'Kazuma').sum()
(df['name'] == 'Kaz').sum()
df[df['name'] == 'Kaz']
model.coef_
pd.Series(model.coef_, index=vectorizer.get_feature_names())
pd.Series(model.coef_, index=vectorizer.get_feature_names()[0])
index = vectorizer.get_feature_names()
index
index.shape
type(index)
pd.Series(model.coef_[0], index=vectorizer.get_feature_names())
coef = pd.Series(model.coef_[0], index=vectorizer.get_feature_names())
coef['Kaz']
coef['az']
coef['azu']
coef['zu']
coef['zum']
coef['uma']
coef['ma']
model.intercept_
coef['z']
coef['u']
coef['m']
coef['a']
coef['Ka']
coef['K']
coef.index
coef['k']
coef['ka']
coef['kaz']

kazvec = vectorizer.transform(['Kazuma'])

np.array(kazvec[kazvec != 0])
np.array(kazvec[kazvec != 0])[0]
coef
coef.values[np.array(kazvec[kazvec != 0])[0] > 0]
coef.values[kazvec != 0]
type(kazvec != 0)
(kazvec != 0).todense()
np.array((kazvec != 0).todense())[0]
coef[np.array((kazvec != 0).todense())[0] != 0]
model.predict(kazvec)
model.predict_proba(kazvec)
pd.Series(model.predict_proba(vectorizer.transform(names))[:, 1], index=names)
model.classes_



from nessvec.indexers import Index
import pandas as pd
from nessvec.files import load_embeddings
df = load_embeddings('word2vec')
small = df.iloc[:300_000]
index = Index(small)
docvector = index.get_doc_vector("This is a planet.")
nearest = index.get_nearest(docvector)
index.get('a')
index.get('a')
nearest
vector
docvector
docvector2 = index.get_doc_vector("This is a planet.", use_idf=True)
nearest2 = index.get_nearest(docvector2)
nearest2
docvector2 = index.get_doc_vector("This is a cold planet.", use_idf=True)
nearest2 = index.get_nearest(docvector2)
nearest2
docvector2 = index.get_doc_vector("This is a far away planet at edge of solar system cold planet.", use_idf=True)
nearest2 = index.get_nearest(docvector2)
nearest2
docvector2 = index.get_doc_vector("This is a far away planet at edge of solar system cold planet asteroid.", use_idf=True)
nearest2 = index.get_nearest(docvector2)
nearest2
hist -o -p -f ch06_extreme_summarization_word2vec_index_use_idf_planet_pluto_asteroid.hist.md
hist -f ch06_extreme_summarization_word2vec_index_use_idf_planet_pluto_asteroid.hist.py



import numpy as np
import pandas as pd
import re
from pathlib import Path

from sklearn.linear_model import LogisticRegression
from pandas.api.types import is_numeric_dtype


# THIS_DIR = Path(__file__).expanduser().resolve().absolute().parent
# DATA_DIR = Path(THIS_DIR).parent.parent / '.nlpia2-data'
DATA_DIR = Path('/home/hobs/code/prosocialai/nlpia2/.nlpia2-data')
DATAFILE = DATA_DIR / 'fake_vs_real.csv.gz'  # 'all.csv.gz'
assert DATAFILE.is_file()


def get_text_column_names(df):
    columns = []
    for c in df.columns:
        try:
            df.head(2)[c].str.len()
            columns.append(c)
        except (AttributeError, ValueError) as e:
            print(e)
            pass
            # print(c + ' is not a str: ' + str(e))
    return columns


def extract_len_features(
        df,
        columns=None,
        pattern=r'\w+(?:\'\w+)?|[^\w\s]',
        pattern_name=None):
    if columns is None:
        columns = get_text_column_names(df)
    for c in columns:
        df[f'{c}_len'] = df[c].str.len()
        df[f'{c}_num_token{pattern_name}s'] = df[c].str.findall(pattern).apply(len)
        df[f'{c}_ave_token{pattern_name}_len'] = (df[f'{c}_num_token{pattern_name}s'] / df[c + '_len'])
        df[f'{c}_ave_token{pattern_name}_len_isna'] = df[f'{c}_ave_token{pattern_name}_len'].isna()
        maximum = df[f'{c}_ave_token{pattern_name}_len'].max()
        if np.isnan(maximum):
            maximum = 1
        df[f'{c}_ave_token{pattern_name}_len'] = df[f'{c}_ave_token{pattern_name}_len'].fillna(maximum)
    return df.fillna(0)


def extract_keyword_features(
        df,
        columns=None,
        keywords='Republican|Trump|Donald|Donald Trump|Senat|CORRUPT|CLINTON|OBAMA|TRUMP'.split('|')):
    if columns is None:
        columns = get_text_column_names(df)
    for c in columns:
        for kw in keywords:
            df[c + '_haskw_' + kw] = df[c].str.contains(kw)
    return df


def train(df, model=LogisticRegression(max_iter=10000), feature_cols=None, target_col=None):
    df = df.sample(len(df))  # <1>
    # <1> Shuffle your data to improve convergence when your data is "stratified" by the target as it was here.

    feature_cols = feature_cols or df.columns.values[:-1]
    target_col = target_col or df.columns.values[-1]

    istestset = np.random.rand(len(df)) < .1  # <1>
    # <1> 10% test dataset for model validation

    df_train = df[~istestset]
    df_test = df[istestset]

    X_train = df_train[feature_cols]
    y_train = df_train[target_col]
    X_test = df_test[feature_cols]
    y_test = df_test[target_col]

    model.fit(X_train, y_train)

    return dict(
        model=model,
        train_score=model.score(X_train, y_train),
        test_score=model.score(X_test, y_test),
        coef=pd.Series(model.coef_.flatten(), index=feature_cols).sort_values())


def tokenize(s):
    pattern = r'\w+(?:\'\w+)?|[^\w\s]'
    return re.findall(pattern, s)


def tokenize_words(s):
    wordpat = r'\w+(?:\'\w+)?'
    return re.findall(wordpat, s)


def filter_nonwords(tokens):
    return [t for t in tokens if t and t[0]]


def _wordlen(tokens):
    return tokenize_words()


def extract_kitchen_sink_features(df, ignore=['subject'], columns=['text', 'title', 'date', 'date_isna', 'isfake']):
    df = df[[c for c in columns if c in df.columns]].copy()
    for i in ignore:
        if i in df.columns:
            del df[i]
    for i, p in enumerate([r'\w+(?:\'\w+)?|[^\w\s]', r'[^\w\s]+', r'[A-Z0-9_]+']):
        df = extract_len_features(df, pattern=p, pattern_name=i)
    # print(f'df.shape before kw features: {df.shape}')
    df = extract_keyword_features(df)
    # print(f'df.shape after kw features: {df.shape}')

    numeric_cols = [c for c in df.columns if is_numeric_dtype(df[c])]
    # delete uniform (constant) numeric columns
    numeric_cols = [c for c in numeric_cols if df[c].std() > 0]

    target_col = 'isfake'
    # feature_cols = [c for c in numeric_cols if c != target_col]
    # print(f'len(numeric_cols): {len(numeric_cols)}')
    # print(f'len(feature_cols): {len(feature_cols)}')
    # print(f'target_col: {target_col}')

    df = df[numeric_cols]
    target = df[target_col].copy()
    del df[target_col]
    df[target_col] = target
    df.to_csv(DATA_DIR / 'numeric.csv.gz', compression='gzip')
    return df


def kitchen_sink_model(df=None, target_col=None):
    if df is None:
        df_filepath = Path(DATA_DIR) / DATAFILE
        df = pd.read_csv(df_filepath, index_col=None)
        print(filepath)
    # print(f'kitchen sink df.shape: {df.shape}')

    df = extract_kitchen_sink_features(df)
    return train(df, target_col=target_col)


def allcaps_tokens(tokens):
    return [tok for tok in tokens if tok.upper() == tok]


def cheap_trick_model(df, target_col='isfake'):
    feature_names = []
    for col in ['text', 'title']:
        feature_names.append(f'{col}_len')
        df[feature_names[-1]] = df[col].apply(len)

        feature_names.append(f'{col}_allcap_token_len')
        df[feature_names[-1]] = df[col].apply(tokenize).apply(allcaps_tokens).apply(len)

        feature_names.append(f'{col}_allcap_ratio')
        df[feature_names[-1]] = df[feature_names[-2]] / df[feature_names[-3]].replace(0, 1)

    return train(df, feature_cols=feature_names, target_col='isfake')


if __name__ == '__main__':
    filepath = Path(DATA_DIR) / DATAFILE
    df = pd.read_csv(filepath, index_col=None)

    results = []
    results.append(cheap_trick_model(df, target_col='isfake'))
    results[-1]['description'] = '## 6 ALLCAPS token features'
    print()
    print(results[-1]['description'])
    print('Test Accuracy: ', results[-1]['test_score'])
    print('### Top 5 isfake indicators:')
    print(results[-1]['coef'].sort_values()[-5:])
    print('### Top 5 NOT fake indicators (negative coefficients):')
    print(results[-1]['coef'].sort_values()[:5])
    print()
    results.append(kitchen_sink_model(df, target_col='isfake'))
    results[-1]['description'] = '## kitchen sink features (keywords and token stats)'
    print()
    print(results[-1]['description'])
    print('Test Accuracy: ', results[-1]['test_score'])
    print('### Top 5 isfake indicators:')
    print(results[-1]['coef'].sort_values()[-5:])
    print('### Top 5 NOT fake indicators (negative coefficients):')
    print(results[-1]['coef'].sort_values()[:5])
    print()
    print('# Hyperparameter Table')
    print(pd.DataFrame(results)['model train_score test_score description'.split()])



"""
FIXME: Verify predict and compute_accuracy() functions by comparing to older versions in git

$ python main.py
WARNING:__main__:args: []
WARNING:__main__:kwargs: {}
100%|████████████████████████████████████████████████████████████████████████████████████████████| 7613/7613 [00:00<00:00, 242018.81it/s]
Epoch: 1, loss: 0.71129, Train accuracy: 0.56970, Test accuracy: 0.64698
Epoch: 2, loss: 0.62170, Train accuracy: 0.64458, Test accuracy: 0.71260
Epoch: 3, loss: 0.52818, Train accuracy: 0.67421, Test accuracy: 0.70997
Epoch: 4, loss: 0.47667, Train accuracy: 0.70690, Test accuracy: 0.73753
Epoch: 5, loss: 0.45808, Train accuracy: 0.73113, Test accuracy: 0.74672
Epoch: 6, loss: 0.40075, Train accuracy: 0.75376, Test accuracy: 0.74803
Epoch: 7, loss: 0.43526, Train accuracy: 0.76558, Test accuracy: 0.75066
Epoch: 8, loss: 0.42900, Train accuracy: 0.79317, Test accuracy: 0.75722
Epoch: 9, loss: 0.49162, Train accuracy: 0.79054, Test accuracy: 0.76247
Epoch: 10, loss: 0.38202, Train accuracy: 0.80324, Test accuracy: 0.75984


$ python main.py --tokenizer=tokenize_spacy --stride=2 --vocab_size=4000 --kernel_lengths=[3,4,5] --text_len=40
WARNING:__main__:args: []
WARNING:__main__:kwargs: {'tokenizer': 'tokenize_spacy', 'stride': '2', 'vocab_size': '4000', 'kernel_lengths': '[3,4,5]', 'text_len': '40'}
WARNING:__main__:NEW KWARGS: tokenizer: tokenize_spacy (<class 'str'>)
WARNING:__main__:NEW KWARGS: vocab_size: 4000 (<class 'int'>)
WARNING:__main__:NEW KWARGS: kernel_lengths: (3, 4, 5) (<class 'tuple'>)
WARNING:__main__:NEW KWARGS: stride: 2 (<class 'int'>)
100%|███████████████████████████████████████████████████████████████████████████████████████████████| 7613/7613 [00:36<00:00, 207.29it/s]
Epoch: 1, loss: 0.73025, Train accuracy: 0.55247, Test accuracy: 0.65748
Epoch: 2, loss: 0.52448, Train accuracy: 0.62808, Test accuracy: 0.68241
Epoch: 3, loss: 0.49091, Train accuracy: 0.67319, Test accuracy: 0.72310
Epoch: 4, loss: 0.42739, Train accuracy: 0.69917, Test accuracy: 0.68898
Epoch: 5, loss: 0.35826, Train accuracy: 0.72179, Test accuracy: 0.74541
Epoch: 6, loss: 0.38785, Train accuracy: 0.74091, Test accuracy: 0.73885
Epoch: 7, loss: 0.38957, Train accuracy: 0.76222, Test accuracy: 0.75066
Epoch: 8, loss: 0.30744, Train accuracy: 0.76850, Test accuracy: 0.75328
Epoch: 9, loss: 0.30294, Train accuracy: 0.78178, Test accuracy: 0.75984
Epoch: 10, loss: 0.39879, Train accuracy: 0.79988, Test accuracy: 0.76115

FIXME:
$ python main.py --tokenizer=tokenize_re --stride=1 --vocab_size=1500 --kernel_lengths=[2,3,4,5] --text_len=35
WARNING:__main__:args: []
WARNING:__main__:kwargs: {'tokenizer': 'tokenize_re', 'stride': '1', 'vocab_size': '1500', 'kernel_lengths': '[2,3,4,5]', 'text_len': '35'}
WARNING:__main__:NEW KWARGS: tokenizer: tokenize_re (<class 'str'>)
WARNING:__main__:NEW KWARGS: vocab_size: 1500 (<class 'int'>)
WARNING:__main__:NEW KWARGS: kernel_lengths: (2, 3, 4, 5) (<class 'tuple'>)
WARNING:__main__:NEW KWARGS: stride: 1 (<class 'int'>)
100%|████████████████████████████████████████████████████████████████████████████████████████████| 7613/7613 [00:00<00:00, 240734.29it/s]
Traceback (most recent call last):
  File "/home/hobs/code/tangibleai/nlpia2/src/nlpia2/ch08/cnn-classify-text/main.py", line 305, in <module>
    pipeline.train()
  File "/home/hobs/code/tangibleai/nlpia2/src/nlpia2/ch08/cnn-classify-text/main.py", line 249, in train
    y_pred = self.model(x_batch)
  File "/home/hobs/anaconda3/envs/nlpia2/lib/python3.9/site-packages/torch/nn/modules/module.py", line 1110, in _call_impl
    return forward_call(*input, **kwargs)
  File "/home/hobs/code/tangibleai/nlpia2/src/nlpia2/ch08/cnn-classify-text/cnn/model.py", line 80, in forward
    x = self.embedding(x)
  File "/home/hobs/anaconda3/envs/nlpia2/lib/python3.9/site-packages/torch/nn/modules/module.py", line 1110, in _call_impl
    return forward_call(*input, **kwargs)
  File "/home/hobs/anaconda3/envs/nlpia2/lib/python3.9/site-packages/torch/nn/modules/sparse.py", line 158, in forward
    return F.embedding(
  File "/home/hobs/anaconda3/envs/nlpia2/lib/python3.9/site-packages/torch/nn/functional.py", line 2183, in embedding
    return torch.embedding(weight, input, padding_idx, scale_grad_by_freq, sparse)
IndexError: index out of range in self


"""

from collections import Counter
from dataclasses import dataclass
from itertools import chain
import logging
from pathlib import Path
import re
import sys
import numpy as np

from sklearn.model_selection import train_test_split
import torch
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from tqdm import tqdm

from cnn.model import CNNTextClassifier
from cnn.language_model import nlp

import pandas as pd

DATA_DIR = Path(__file__).parent / 'data'

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.WARNING)


def tokenize_spacy(doc):
    return [tok.text for tok in nlp(doc) if tok.text.strip()]


def tokenize_re(doc):
    return [tok for tok in re.findall(r'\w+', doc)]


@dataclass(init=True, repr=True, eq=True, order=False, unsafe_hash=False, frozen=False)
class Parameters:
    filepath: str = 'disaster-tweets.csv'
    usecols: tuple = ('text', 'target')
    tokenizer: str = 'tokenize_re'

    seq_len: int = 35
    vocab_size: int = 2000

    embedding_size: int = 64
    kernel_lengths: tuple = (2, 3, 4, 5)
    strides: tuple = (2, 2, 2, 2)
    conv_output_size: int = 32
    stride: int = 2

    # Training parameters
    epochs: int = 10
    batch_size: int = 12
    learning_rate: float = 0.001
    test_size = 0.1

    dropout_portion = 0.2

    case_sensitive = True


HYPERPARAMS = Parameters()


def pad(sequence, pad_value=0, seq_len=HYPERPARAMS.seq_len):
    log.debug(f'BEFORE PADDING: {sequence}')
    padded = list(sequence)[:seq_len]
    padded = padded + [pad_value] * (seq_len - len(padded))
    log.debug(f'AFTER PADDING: {sequence}')
    return padded


def load_dataset_spacy(filepath='tweets.csv', usecols=[0, -1], tokenizer=tokenize_spacy, params=HYPERPARAMS):
    """ load and preprocess csv file: return [(token id sequences, label)...]

    1. Simplified: load the CSV
    2. NOPE: case folding:
    3. NOPE: remove non-letters (nonalpha):
    4. NOPE: remove stopwords
    5. Simplified: tokenize with regex
    6. Simplified: filter infrequent words
    7. Simplified: compute reverse index
    8. Simplified: transform token sequences to integer id sequences
    9. Simplified: pad token id sequences
    10. Simplified: train_test_split
    """

    if not Path(filepath).is_file():
        filepath = DATA_DIR / filepath

    # 1. Simplified: load the CSV

    df = pd.read_csv(filepath, usecols=list(usecols))

    # 2. NOPE: case folding:

    texts = map(str.lower, df['texts'])

    # 3. NOPE: remove non-letters (nonalpha):
    # texts = re.sub(r'[^A-Za-z]', t, ' ') for t in texts]
    # 4. NOPE: remove stopwords
    # 5. Simplified: tokenize with regex

    tokenized_texts = map(re.compile(r'\w+').findall, texts)

    # 6. Simplified: filter infrequent words

    counts = Counter(chain(*tokenized_texts))
    vocab = ['<PAD>'] + [x[0] for x in counts.most_common(params.vocab_size)]

    # 7. Simplified: compute reverse index

    tok2id = dict(zip(vocab, range(len(vocab))))

    # 8. Simplified: transform token sequences to integer id sequences

    id_sequences = [map(tok2id.get, seq) for seq in tokenized_texts]

    # 9. Simplified: pad token id sequences

    id_sequences = [list(map(pad, seq)) for seq in id_sequences]

    # 10. Simplified: train_test_split

    return dict(zip(
        'x_train x_test y_train y_test'.split(),
        train_test_split(
            X=id_sequences,
            y=df['target'],
            test_size=params.test_size,
            random_state=0)))


def load_dataset(filepath='tweets.csv', usecols=[0, -1], tokenizer=tokenize_re, params=HYPERPARAMS, **kwargs):
    """ load and preprocess csv file: return [(token id sequences, label)...]

    1. Simplified: load the CSV
    2. NOPE: case folding:
    3. NOPE: remove non-letters (nonalpha):
    4. NOPE: remove stopwords
    5. Simplified: tokenize with regex
    6. Simplified: filter infrequent words
    7. Simplified: compute reverse index
    8. Simplified: transform token sequences to integer id sequences
    9. Simplified: pad token id sequences
    10. Simplified: train_test_split
    """
    if not Path(filepath).is_file():
        filepath = Path(DATA_DIR) / filepath
    df = pd.read_csv(filepath.open(), usecols=usecols)
    texts = df[usecols[0]].values
    targets = df[usecols[1]].values
    texts = [re.sub(r'[^A-Za-z0-9.?!]+', ' ', x) for x in texts]
    texts = [tokenizer(doc) for doc in tqdm(texts)]
    counts = Counter(chain(*texts))
    vocab = ['<PAD>'] + [x[0] for x in counts.most_common(params.vocab_size)]
    tok2id = dict(zip(vocab, range(len(vocab))))

    # 8. Simplified: transform token sequences to integer id sequences

    id_sequences = [[i for i in map(tok2id.get, seq) if i is not None] for seq in texts]

    # 9. Simplified: pad token id sequences

    padded_sequences = []
    for s in id_sequences:
        padded_sequences.append(pad(s))
    padded_sequences = torch.IntTensor(padded_sequences)

    return dict(zip(
        'x_train x_test y_train y_test'.split(),
        train_test_split(
            padded_sequences,
            targets,
            test_size=HYPERPARAMS.test_size,
            random_state=0)))


class DatasetMapper(Dataset):

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


def calculate_accuracy(y_true, y_pred):
    y_true, y_pred = np.array(y_true), np.array(y_pred)
    true_positives = 0
    true_negatives = 0
    for true, pred in zip(y_true, y_pred):
        if (pred >= 0.5) and (true == 1):
            true_positives += 1
        elif (pred < 0.5) and (true == 0):
            true_negatives += 1
        else:
            pass
    # Return accuracy
    return (true_positives + true_negatives) / len(y_true)


class Pipeline(Parameters):

    def __init__(self, *args, **kwargs):
        # self.x_train, self.y_train, self.x_test, self.y_test:
        super().__init__()
        log.info(args)
        log.info(kwargs)
        for param_name, param_val in super().__dict__.items():

            log.info(f'DEFAULT: {param_name}: {param_val}')
            kwarg_val = kwargs.get(param_name)
            if kwarg_val is not None:
                dest_type = type(param_val)
                if not isinstance(param_val, str) and isinstance(kwarg_val, str):
                    kwarg_val = eval(kwarg_val)
                setattr(self, param_name, dest_type(kwarg_val))
                log.warning(f'NEW KWARGS: {param_name}: {getattr(self, param_name)} ({type(getattr(self, param_name))})')
        self.__dict__.update(load_dataset(
            filepath=self.filepath,
            usecols=list(self.usecols),
            tokenizer=globals()[self.tokenizer],
            **kwargs))
        model_kwargs = {k: v for (k, v) in vars(self).items() if not k.startswith('_')}
        log.debug(f'MODEL_KWARGS: {model_kwargs}')
        self.model = CNNTextClassifier(params=HYPERPARAMS, **model_kwargs)

    def train(self):

        self.trainset_mapper = DatasetMapper(self.x_train, self.y_train)
        self.testset_mapper = DatasetMapper(self.x_test, self.y_test)

        self.loader_train = DataLoader(self.trainset_mapper, batch_size=self.batch_size)
        self.loader_test = DataLoader(self.testset_mapper, batch_size=self.batch_size)

        optimizer = optim.RMSprop(self.model.parameters(), lr=self.learning_rate)

        for epoch in range(self.epochs):
            self.model.train()
            predictions = []
            for x_batch, y_batch in self.loader_train:
                y_batch = y_batch.type(torch.FloatTensor)
                y_pred = self.model(x_batch)
                loss = F.binary_cross_entropy(y_pred, y_batch)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                # Save predictions
                predictions += list(y_pred.detach().numpy())

            test_predictions = self.predict()
            train_accuary = calculate_accuracy(self.y_train, predictions)
            test_accuracy = calculate_accuracy(self.y_test, test_predictions)
            print("Epoch: %d, loss: %.5f, Train accuracy: %.5f, Test accuracy: %.5f" % (epoch + 1, loss.item(), train_accuary, test_accuracy))
        return self

    def predict(self, X=None):

        self.model.eval()  # evaluation mode
        predictions = []

        if X is not None:
            X_batches = zip(X, [None] * len(X))
        else:
            X_batches = list(zip(*self.loader_test))[0]
        with torch.no_grad():
            for x_batch in X_batches:
                y_pred = self.model(x_batch)
                predictions += list(y_pred.detach().numpy())

        return predictions

    def score(self, X, y):
        y_pred = self.predict(X)
        return torch.Tensor.mean((y_pred - y)**2) ** .5


def parse_argv(sys_argv=sys.argv):
    argv = list(reversed(sys_argv[1:]))

    pipeline_args = []
    pipeline_kwargs = dict(tokenizer='tokenize_re')
    while len(argv):
        a = argv.pop()
        if a.startswith('--'):
            if '=' in a:
                k, v = a.split('=')
                k = k.lstrip('-')
            else:
                k = a.lstrip('-')
                v = argv.pop()
            pipeline_kwargs[k] = v
        else:
            pipeline_args.append(a)

    return pipeline_args, pipeline_kwargs


def main():

    pipeline_args, pipeline_kwargs = parse_argv(sys.argv)

    log.warning(f'args: {pipeline_args}')
    log.warning(f'kwargs: {pipeline_kwargs}')

    pipeline = Pipeline(*pipeline_args, **pipeline_kwargs)

    pipeline = pipeline.train()
    predictions = pipeline.predict()

    return dict(pipeline=pipeline, predictions=predictions)


if __name__ == '__main__':
    import json
    results = main()
    json.dump(
        [float(x) for x in results['predictions']],
        open('predictions.json', 'a'), indent=2)



from nltk.tokenize import sent_tokenize, word_tokenize
import syllables

def readability_ease_score(text):
    """Return the Flesch readability ease score of a text.
    Interpretation: the more difficult the text is, the lower the score.
    
    >>> readability_ease_score("data science for social good")
    75.88
    >>> readability_ease_score("what time is it?")
    117.16
    >>> readability_ease_score(nan)
    nan
    """
    
    if str(text) != 'nan':
        num_sentences = len(sent_tokenize(text))
        num_words = len(word_tokenize(text))
        num_syllables = sum(syllables.estimate(w) for w in word_tokenize(text))
        score = 206.835 - 1.015 * (float(num_words) / num_sentences) - 84.6 * (num_syllables / float(num_words))
        return round(score, 2)
    else:
        return np.nan



import pandas as pd
import requests

url = 'https://www.ethnologue.com/sites/default/files/LanguageCodes.tab'

# urllib can't GET the page, but requests.get has not problem:
# pd.read_csv(url, sep='\t')

r = requests.get(url)
data = [line.split('\t') for line in r.text.splitlines()]
columns = data[0]
df = pd.DataFrame(data[1:], columns=columns)
df
#      LangID CountryID LangStatus                 Name
# 0       aaa        NG          L               Ghotuo
# 1       aab        NG          L           Alumu-Tesu
# 2       aac        PG          L                  Ari
# 3       aad        PG          L                 Amal
# 4       aae        IT          L  Albanian, Arbëreshë
# ...     ...       ...        ...                  ...
# 7481    zyg        CN          L         Zhuang, Yang
# 7482    zyj        CN          L     Zhuang, Youjiang
# 7483    zyn        CN          L      Zhuang, Yongnan
# 7484    zyp        MM          L          Chin, Zyphe
# 7485    zzj        CN          L     Zhuang, Zuojiang

# [7486 rows x 4 columns]




import spacy

LANG_MODEL_NAME = 'en_core_web_md'

nlp = None
for i in range(2):
    try:
        nlp = spacy.load(LANG_MODEL_NAME)
    except IOError:
        spacy.cli.download(LANG_MODEL_NAME)
    if nlp is not None:
        break



import os
import subprocess
import re
import pandas as pd
from pathlib import Path
import argparse

DIRS 		= ['nlpia-manuscript', 'nlpia2']
DF_COLUMNS  = ['Module', 'Line', 'Line #', 'Path', 'Repo Root', 'Pattern']

## IDENTIFY GIT REPO ROOT
# This ensures the script will work even if it's moved around within the repository
REPO_COMMAND = subprocess.run(['git', 'rev-parse', '--show-toplevel'], capture_output=True)
CURRENT_REPO_ROOT =Path( REPO_COMMAND.stdout.strip(b'\n').decode('utf-8') )

REPOS = [ (CURRENT_REPO_ROOT.parent / directory) for directory in DIRS]
MANUSCRIPT_ADOCS = CURRENT_REPO_ROOT.parent / 'nlpia-manuscript' / 'manuscript' / 'adoc'


## REGULAR EXPRESSION PATTERNS
FROM_R 		= r'[>\s]*from\s+(?P<from_import>\w+[\w\.]*)\s+import\s+.*#*.*'
#https://regex101.com/r/afZx6f/1

IMPORT_AS_R = r'[>\s]*import\s+(?P<import_as>\w+)\.?\w*\s+(\sas\s)?.*#*.*'
#https://regex101.com/r/ALfXkh/1

IMPORT_R 	= r'[>\s]*import\s+(?P<import_direct>(\w+[,\s*]?)*)\s*[#]*.*'
#https://regex101.com/r/x2gxek/1

PATTERNS_R 	= (FROM_R, IMPORT_AS_R, IMPORT_R)


def inspect_files(root, pattern, expanded=False):
	"""
	Expects a local path and a pattern of documents for glob
	Inspects all the files that match the pattern, and extracts the import statements into a dataframe.
	Prints on screen a set of all imported modules found
	"""
	print('\n'*3, '*'*30, str(root), sep="")
	root=str(root)
	import_lines=[]
	terms_rc = re.compile(r'|'.join(PATTERNS_R))
	for filepath in Path(root).rglob(pattern):
		with open(filepath) as f:
			for number, line in enumerate(f):
				line = line.strip('\n ')
				m = terms_rc.match(line) 
				if m:
					modules = m.group('from_import') or m.group('import_as') or m.group('import_direct') 
					import_lines.append( (modules, line,  number+1, str(filepath).split(root)[1], root, pattern) )
	modules_df =  pd.DataFrame(import_lines, columns = DF_COLUMNS)
	modules_df['Module'] = modules_df['Module'].apply(lambda x: x.split('.')[0] if '.' in x else x)
	modules_df['Module'] = modules_df['Module'].apply(lambda x: [y.strip() for y in x.split(',')] if ',' in x else x)
	modules_df = modules_df.explode(['Module'], ignore_index=True)
	if expanded==True:
		with pd.option_context('display.max_rows', None, 'display.max_columns', None, 'display.max_colwidth', None):
			print( modules_df.groupby(by=['Module', 'Path', 'Line #', 'Line'])['Line'].agg('count'))
	print('\nModules identified, with counts: ')
	print ( modules_df['Module'].value_counts())
	
	print("\nList of Modules: \n\t", end="")
	print( * sorted(set( modules_df['Module'])), sep = ", ")
	
	return modules_df


if __name__ == '__main__':

	parser = argparse.ArgumentParser(description="%(prog)s recursively traverses repositories or subrepo directory structures, \
		inspecting files whose names match a given name pattern, to collect package names that appear in import statements. The output \
		lists such packages along with their counts. With the optional --expand flag, %(prog)s additionally outputs the line numbers and filenames where \
		the imports were found.")
	parser.add_argument("-a", "--all", dest="all", action='store_true',
		default = False, 
		help="exhaustively traverse both repos (nlpia-manuscript, nlpia2)")
	parser.add_argument("-p", "--pattern", dest="pattern", metavar="",
		default = "*.adoc",
		help="pattern of filenames for glob to match, default is '*.adoc'" )
	parser.add_argument("-e", "--expand", dest="expand", action='store_true',
		default=False,
		help="additionally prints each occurence of a package mention, along with file name and line number", )
	args = parser.parse_args()

	if args.all == False:													# checks only the manuscript_adocs directory
		repo_root = MANUSCRIPT_ADOCS
		df = inspect_files(repo_root, args.pattern, args.expand)
	else:																	# otherwise, check all repos in the DIRS list
		for repo_root in REPOS: 
			df = inspect_files(repo_root, args.pattern, args.expand)



import pandas as pd
import numpy as np

import qary
import qary.etl.faqs

df_dict = qary.etl.faqs.load(domains='data')
df = pd.DataFrame(np.array([df_dict['questions'], df_dict['answers']]).T)
df.columns = 'question answer'.split()
df = df[df.question.str.endswith('?')].copy()

df.to_csv('all-faqs.csv', index=False)



import pandas as pd
from pathlib import Path
DATA_DIR = Path('~/code/tangibleai/fake-news/data').expanduser().resolve().absolute()
df_fake = pd.read_csv(Path(DATA_DIR) / 'Fake.csv', index_col=None)
df_true = pd.read_csv(Path(DATA_DIR) / 'True.csv', index_col=None)
df_fake['isfake'] = 1
df_true['isfake'] = 0
df = pd.concat([df_true, df_fake], ignore_index=True)
"""
>>> import pandas as pd
... from pathlib import Path
... DATA_DIR = Path('~/code/tangibleai/fake-news/data').expanduser().resolve().absolute()
... df_fake = pd.read_csv(Path(DATA_DIR) / 'Fake.csv', index_col=None)
... df_true = pd.read_csv(Path(DATA_DIR) / 'True.csv', index_col=None)
... df_fake['isfake'] = 1
... df_true['isfake'] = 0
... df = pd.concat([df_true, df_fake], ignore_index=True)
>>> df.describe(include='all')
                                                    title   text       subject                date        isfake
count                                               44898  44898         44898               44898  44898.000000
unique                                              38729  38646             8                2397           NaN
top     Factbox: Trump fills top jobs for his administ...         politicsNews  December 20, 2017            NaN
freq                                                   14    627         11272                 182           NaN
mean                                                  NaN    NaN           NaN                 NaN      0.522985
std                                                   NaN    NaN           NaN                 NaN      0.499477
min                                                   NaN    NaN           NaN                 NaN      0.000000
25%                                                   NaN    NaN           NaN                 NaN      0.000000
50%                                                   NaN    NaN           NaN                 NaN      1.000000
75%                                                   NaN    NaN           NaN                 NaN      1.000000
max                                                   NaN    NaN           NaN                 NaN      1.000000
"""

df['date'] = pd.to_datetime(df['date'], errors='coerce')
"""
>>> df['date'].isna().sum()
10
"""

df['date_isna'] = df['date'].isna()
df['date'] = df['date'].fillna('mean')
"""
>>> df['date'].isna().sum()
0
"""

df.to_csv(Path(DATA_DIR) / 'all.csv.gz', index=False)



from ch05.fake_news_logistic_regression_title_len import *
import numpy as np

try:
    assert len(__file__)
except (AssertionError, NameError):
    __file__ = "fake_news_logistic_regression_title_len_red_blue.py"

data = df[['title_len', 'isfake']].sample(250)
data['proba'] = model.predict_proba(data[['title_len']])[:, 1]
data = data.sort_values('proba')
isfake = data['isfake'] > 0
above50 = data['proba'] > .5

# ax = sns.regplot(x=data['title_len'][~isfake], y=data['isfake'][~isfake], logistic=True, color=color)
ax = sns.scatterplot(
    x=data['title_len'][~isfake], y=data['isfake'][~isfake],
    s=60, color='darkblue', alpha=.4)
ax = sns.scatterplot(
    x=data['title_len'][isfake], y=data['isfake'][isfake],
    s=60, color='darkred', alpha=.45)
# sns.lineplot(x=data['title_len'][isfake], y=data['title_len'][isfake].apply(estimator), color='red')
color = np.array(['r'] * len(data))


sns.lineplot(
    x=data['title_len'][~above50], y=data['proba'][~above50],
    linewidth=4, color='darkblue', alpha=.4, legend=False)
sns.lineplot(
    x=data['title_len'][above50], y=data['proba'][above50],
    linewidth=4, color='darkred', alpha=.45, legend=False)
plt.xlabel('Title Length (chars)')
plt.ylabel('Probability of Being Fake News')

image_filepath = CH05_IMAGES_DIR / (Path(__file__).name[:-2] + 'png')
print(f'Saving logistic regression plot to {image_filepath}')
plt.savefig(image_filepath)
plt.show()




url = 'https://www.youtube.com/watch?v=snkwsU98QlQ'
url = 'https://upload.wikimedia.org/wikipedia/commons/d/de/Hello_in_Morse_Code.wav'
url = 'https://commons.wikimedia.org/wiki/File:1210secretmorzecode.wav'

import requests
requests.get?
wav = requests.get(url)
wav.content
url = 'https://upload.wikimedia.org/wikipedia/commons/7/78/1210secretmorzecode.wav'
wav = requests.get(url)
wav
wav.content
headers = {'User-Agent': 'NLPiA2Bot/0.7 (https://tangibleai.com/about/; nlpia2@tangibleai.com)'}
wav = requests.get(url, headers=headers)
wav
wav.content
bytearray(wav.content)
resp = wav
wav = pd.Series([b for b in resp.content])
import pandas as pd
wav = pd.Series([b for b in resp.content])
wav
wav = pd.Series(resp.content)
wav
chr(82
    )
from urllib.request import urlretrieve
f = urlretrieve(url, '1210secretmorzecode.wav')
ls - hal
from scipy.io import wavfile
samplerate, data = wavfile.read('1210secretmorzecode.wav')
data
wav = pd.Series(data)
wavfile?
dir(wavfile)
dir(wavfile)
wav = pd.Series(data)
wav.plot()
from seaborn import pyplot as plt
from matplotlib import pyplot
from matplotlib import pyplot as plt
plt.show()
wav.describe()
pwd
wav
from scipy.io import wavfile
samplerate, data = wavfile.read('1210secretmorzecode.wav')
samplerate
data
data[0]
data.shape
import wave
import struct

wavefile = wave.open('sine.wav', 'r')

length = wavefile.getnframes()
for i in range(0, length):
    wavedata = wavefile.readframes(1)
    data = struct.unpack("<h", wavedata)
    print(int(data[0]))
filename = '1210secretmorzecode.wav'
import wave
import struct

wavefile = wave.open(filename, 'r')

length = wavefile.getnframes()
for i in range(0, length):
    wavedata = wavefile.readframes(1)
    data = struct.unpack("<h", wavedata)
    print(int(data[0]))
wavedata
import wave
import struct

wavefile = wave.open(filename, 'r')

length = wavefile.getnframes()
for i in range(0, length):
    wavedata = wavefile.readframes(2)
    data = struct.unpack("<h", wavedata)
    print(int(data[0]))
!pip install SoundFile
sf = SoundFile(filename)
import SoundFile as sf
import soundfile as sf
sf.read(filename)
wav = sf.read(filename)
wav.shape
wav
wav[0]
wav[0].shape
wav[0]
wav[0][1:2:]
wav[0][1::2]
wav[0][0::2]
pd.Series(wav[0][0::2]).plot()
plt.show()
pd.Series(wav[0][0::2][:10000]).plot()
plt.show()
pd.Series(wav[0][:10000]).plot()
plt.show()
wavedata
wav = pd.Series(resp.content)
wav
from scipy.io import wavfile
samplerate, data = wavfile.read('1210secretmorzecode.wav')
wav = pd.Series(data)
wav = pd.Series(data[:5000])
wav.plot()
plt.show()
wav = pd.Series(data[:10000])
wav.plot()
plt.show
plt.show()
samplerate
wav.index = wav.index / samplerate
wav.plot()
plt.show()
plt.show()
wav
wav.plot()
plt.show()
wav.plot()
plt.show()
pwd
hist - o - p - f code / tangibleai / nlpia2 / src / nlpia2 / ch07 / download_read_plot_wav_file.hist.md
hist - f code / tangibleai / nlpia2 / src / nlpia2 / ch07 / download_read_plot_wav_file.hist.py



import random
from annoy import AnnoyIndex
from nessvec.util import load_glove
from tqdm import tqdm


def find_similar(w, n=10):
    return [i2w[i] for i in aidx.get_nns_by_item(w2i['seattle'], 10)]


try:
    w2v['the']
except:
    w2v = load_glove()

class VectorIndex():

    def __init__(self, w2v=w2v, num_trees=10):
        self.idx = AnnoyIndex(len(w2v['the']), 'angular')

        print(f'Building index on {len(w2v)} {len(w2v[w2v.keys()[0]])}-D vectors')
        for i, v in tqdm(enumerate(w2v.values())):
            self.idx.add_item(i, v)

        self.idx.build(num_trees)
        self.i2w = list(w2v.keys())
        self.w2i = dict(zip(i2w, range(len(i2w))))

        return dict(aidx=aidx, i2w=i2w, w2i=w2i)

    def find_similar(w, n=10):
        return [i2w[i] for i in aidx.get_nns_by_item(w2i['seattle'], 10)]


if __name__ == "__main__":
    seattle = w2v['seattle']
    d = create_index()
    aidx = d['aidx']
    aidx.save(f'glove{len(w2v["the"])}.ann')



import pandas as pd


## Load Data
DATA_DIR = ('https://gitlab.com/prosocialai/nlpia2/-/raw/main/.nlpia2-data')

true = pd.read_csv(DATA_DIR+'/fake-news-dataset-true-small.csv')
fake = pd.read_csv(DATA_DIR+'/fake-news-dataset-fake-small.csv')

true['label'] = 1
fake['label'] = 0

fake.drop(labels=['subject','date', 'text'],axis=1,inplace=True)
true.drop(labels=['subject','date', 'text'],axis=1,inplace=True)

df = pd.concat([fake,true])

data.head()

## Clean the text


import pandas as pd
pd.options.display.width = 120  # <1>
pd.set_option('display.max_columns', 7)

DATA_DIR = ('https://gitlab.com/tangibleai/nlpia/-/raw/master/src/nlpia/data')
url= DATA_DIR + '/toxic_comment_small.csv'
comments = pd.read_csv(url)
index = ['comment{}{}'.format(i, '!'*j) for (i,j) in zip(range(len(comments)), comments.toxic)]  # <2>
comments = pd.DataFrame(comments.values, columns=comments.columns, index=index)
mask = comments.toxic.astype(bool).values
comments['toxic'] = comments.toxic.astype(int)
"""
>>> comments.head(6)
                                                        text  toxic
comment0   you have yet to identify where my edits violat...      0
comment1   "\n as i have already said,wp:rfc or wp:ani. (...      0
comment2   your vote on wikiquote simple english when it ...      0
comment3   your stalking of my edits i've opened a thread...      0
comment4!  straight from the smear site itself. the perso...      1
comment5   no, i can't see it either - and i've gone back...      0
"""
#Preprocessing




from sklearn.feature_extraction.text import TfidfVectorizer
import spacy

nlp = spacy.load("en_core_web_sm")

def spacy_tokenizer(sentence):
    return [token.text for token in nlp(sentence.lower())]

tfidf = TfidfVectorizer(tokenizer=spacy_tokenizer)
tfidf_docs = tfidf.fit_transform(raw_documents=comments.text).toarray()
"""
>>> tfidf_docs.shape
(5000, 25172)
>>> comments.toxic.sum()
650
"""

mask = comments.toxic.astype(bool).values  # <1>
toxic_centroid = tfidf_docs[mask].mean(axis=0) # <2>
nontoxic_centroid = tfidf_docs[~mask].mean(axis=0)

toxicity_score = tfidf_docs.dot(toxic_centroid - nontoxic_centroid)
"""
>>> toxicity_score
array([-0.01469806, -0.02007376,  0.03856095, ..., -0.01014774, -0.00344281,  0.00395752])
"""

from sklearn.preprocessing import MinMaxScaler
comments['manual_score'] = MinMaxScaler().fit_transform(toxicity_score.reshape(-1, 1))
comments['manual_predict'] = (comments.manual_score > .5).astype(int)
"""
>>> comments['toxic manual_predict manual_score'.split()].round(2).head(6)
           toxic  manual_predict  manual_score
comment0       0               0          0.41
comment1       0               0          0.27
comment2       0               0          0.35
comment3       0               0          0.47
comment4!      1               0          0.48
comment5       0               0          0.31

"""

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
lda_tfidf = LinearDiscriminantAnalysis()
lda_tfidf = lda_tfidf.fit(tfidf_docs, comments['toxic'])  # <1>
comments['tfidf_predict'] = lda_tfidf.predict(tfidf_docs)
"""
round(float(lda_tfidf.score(tfidf_docs, comments['toxic'])), 3)
0.999
"""

from sklearn.model_selection import train_test_split
X_train, X_test, y_train, y_test = train_test_split(tfidf_docs, \
                comments.toxic.values, test_size=0.5, random_state=271828) # <1>
lda_tfidf_train = LinearDiscriminantAnalysis(n_components=1)
lda_tfidf_train = lda_tfidf_train.fit(X_train, y_train)  # <2>
"""
>>> round(float(lda_tfidf_train.score(X_train, y_train)), 3)
0.99
>>> round(float(lda_tfidf_train.score(X_test, y_test)), 3)
0.554
"""

from sklearn.metrics import confusion_matrix
confusion_matrix(y_test, lda_tfidf_train.predict(X_test))
"""
array([[1261,  913],
       [ 201,  125]], dtype=int64)
"""

import matplotlib.pyplot as plt
from sklearn.metrics import plot_confusion_matrix
plot_confusion_matrix(lda_tfidf_train,X_test, y_test, cmap="Greys",
                      display_labels=['non-toxic', 'toxic'], colorbar=False)
#plt.show()

from sklearn.metrics import f1_score
f1_score(y_test,lda_tfidf_train.predict(X_test))
"""
0.1832844574780059
"""


### Latent Semantic Analysis
from sklearn.decomposition import TruncatedSVD
tfidf_docs_centered = tfidf_docs - tfidf_docs.mean()
svd = TruncatedSVD(n_components=16, n_iter=100)  # <1>
columns = ['topic{}'.format(i) for i in range(svd.n_components)]
svd_topic_vectors = svd.fit_transform(tfidf_docs)
svd_topic_vectors = pd.DataFrame(svd_topic_vectors, columns=columns,
                  index=index)
"""
>>> svd_topic_vectors.round(3).head(6)
           topic0  topic1  topic2  topic3  ...  topic12  topic13  topic14  topic15
comment0    0.121  -0.055   0.036  -0.040  ...    0.013   -0.038    0.089    0.011
comment1    0.215   0.141  -0.006  -0.006  ...   -0.040    0.079   -0.016   -0.070
comment2    0.342  -0.200   0.044  -0.070  ...    0.059   -0.138    0.023    0.069
comment3    0.130  -0.074   0.034  -0.018  ...    0.119   -0.060    0.014    0.073
comment4!   0.166  -0.081   0.040   0.136  ...    0.066   -0.008    0.063   -0.020
comment5    0.256  -0.122  -0.055   0.082  ...    0.011    0.093   -0.083   -0.074

"""

"""
>>> list(tfidf_model.vocabulary_.items())[:5] #<1>
[('you', 18890),
 ('have', 8093),
 ('yet', 18868),
 ('to', 17083),
 ('identify', 8721)]
"""

column_nums, terms = zip(*sorted(zip(tfidf.vocabulary_.values(),
     tfidf.vocabulary_.keys())))  # <2>
"""
>>> terms
('\n', '\n ', '\n \n', '\n \n ', '\n  ')
"""
topic_term_matrix = pd.DataFrame(svd.components_, columns=terms,
                   index=['topic{}'.format(i) for i in range(16)])
"""
>>> pd.options.display.max_columns = 8
>>> topic_term_matrix.head(4).round(3)
>>> toxic_terms= topic_term_matrix['pathetic crazy stupid lazy idiot hate die kill'.split()].round(3) * 100
>>> 

"""

X_train_16d, X_test_16d, y_train_16d, y_test_16d = train_test_split(svd_topic_vectors, \
                                                    comments.toxic.values, test_size=0.5, random_state=271828)
lda_lsa = LinearDiscriminantAnalysis(n_components=1)
lda_lsa = lda_lsa.fit(X_train_16d, y_train_16d)  # <2>
"""
>>> round(float(lda_lsa.score(X_train_16d, y_train_16d)), 3)
0.881
>>> round(float(lda_lsa.score(X_test_16d, y_test_16d)), 3)
0.88
"""
from sklearn.metrics import f1_score
f1_score(y_test_16d, lda_lsa.predict(X_test_16d))
"""
# comparing to PCA 
from sklearn.decomposition import PCA
pca_model = PCA(n_components=16)
tfidf_docs_16d = pca_model.fit_transform(tfidf_docs)


from sklearn.model_selection import train_test_split
X_train_16d, X_test_16d, y_train_16d, y_test_16d = train_test_split(tfidf_docs_16d, \
                                                    comments.toxic.values, test_size=0.5, random_state=271828)
lda_lsa = LinearDiscriminantAnalysis(n_components=1)
lda_lsa = lda_lsa.fit(X_train_16d, y_train_16d)  # <2>
round(float(lda_lsa.score(X_train_16d, y_train_16d)), 3)
round(float(lda_lsa.score(X_test_16d, y_test_16d)), 3)
"""

# Hyperparameter table
hparam_table = pd.DataFrame()

tfidf_performance = {'classifier': 'LDA',
                     'features': 'tf-idf (spacy tokenizer)',
                     'train_accuracy': 0.99 ,
                     'test_accuracy': 0.554,
                     'test_precision': 0.383 ,
                     'test_recall': 0.12,
                     'test_f1': 0.183}

hparam_table = hparam_table.append(tfidf_performance, ignore_index=True)

from sklearn.metrics import precision_score, recall_score, f1_score

def hparam_rec(model, X_train, y_train, X_test, y_test, classifier_name, features):
    return {'classifier': classifier_name,
            'features': features,
            'train_accuracy': float(model.score(X_train, y_train)),
            'test_accuracy': float(model.score(X_test, y_test)),
            'test_precision': precision_score(y_test, model.predict(X_test)),
            'test_recall': recall_score(y_test, model.predict(X_test)),
            'test_f1': f1_score(y_test, model.predict(X_test)) }

lsa_performance = hparam_rec(lda_lsa, X_train_16d, y_train_16d, X_test_16d,y_test_16d, 'LDA', 'LSA (16d)')
hparam_table = hparam_table.append(lsa_performance,ignore_index=True)

def evaluate_model(X,y, classifier, classifier_name, features):
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.5, random_state=271828)
    classifier = classifier.fit(X_train, y_train)
    return hparam_rec(classifier, X_train, y_train, X_test,y_test,
                      classifier_name, features)

from sklearn.feature_extraction.text import CountVectorizer

counter = CountVectorizer(tokenizer=spacy_tokenizer)
counter = counter.fit(comments.text)

bow_docs = pd.DataFrame(counter.transform(comments.text)
                        .toarray(), index=index)
column_nums, terms = zip(*sorted(zip(counter.vocabulary_.values(),
                                     counter.vocabulary_.keys())))
bow_docs.columns = terms

"""
>>> comments.loc['comment0'].text
"""


from sklearn.decomposition import LatentDirichletAllocation as LDiA
ldia = LDiA(n_components=16, learning_method='batch')
ldia = ldia.fit(bow_docs)  # <1>
ldia.components_.shape
columns = ['topic{}'.format(i) for i in range(16)]
ldia16_topic_vectors = ldia.transform(bow_docs)
ldia16_topic_vectors = pd.DataFrame(ldia16_topic_vectors,
                                    index=index, columns=columns)

components = pd.DataFrame(ldia.components_.T, index=terms,
    columns=columns)
components.round(2).head(3)



from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
model_ldia16 = LinearDiscriminantAnalysis()
ldia16_performance =evaluate_model(ldia16_topic_vectors, comments.toxic, model_ldia16, 'LDA', 'LDIA (16d)')

hparam_table = hparam_table.append(ldia16_performance, ignore_index = True)

ldia32 = LDiA(n_components=32, learning_method='batch')
ldia32_topic_vectors = ldia.fit_transform(bow_docs)
model_ldia32 = LinearDiscriminantAnalysis()
ldia32_performance =evaluate_model(ldia32_topic_vectors, comments.toxic, model_ldia32, 'LDA', 'LDIA (32d)')

hparam_table = hparam_table.append(ldia32_performance, ignore_index = True)

words = counter.get_feature_names()

REPO_URL = 'https://gitlab.com/tangibleai/qary/-/raw/master'
FAQ_DIR = 'src/qary/data/faq'
FAQ_FILENAME = 'short-faqs.csv'
DS_FAQ_URL = '/'.join([REPO_URL, FAQ_DIR, FAQ_FILENAME])

df = pd.read_csv(DS_FAQ_URL)

vectorizer = TfidfVectorizer()
df['questions+answers'] = df['question'] + df['answer']
vectorizer.fit(df['questions+answers'])
# vectorize all the questions/answers in qa_dataset
tfidfvectors_sparse = vectorizer.transform(df['question'])
tfidfvectors = tfidfvectors_sparse.todense()
svd = TruncatedSVD(n_components=16, n_iter=100)
tfidfvectors_16d = svd.fit_transform(tfidfvectors)

def bot_reply(question):
      question_tfidf = vectorizer.transform([question]).todense()
      question_16d = svd.transform(question_tfidf)
      idx = question_16d.dot(tfidfvectors_16d.T).argmax()
      print(
            f"Your question:\n  {question}\n\n"
            f"Most similar FAQ question:\n  {df['question'][idx]}\n\n"
            f"Answer to that FAQ question:\n  {df['answer'][idx]}\n\n"
           )

bot_reply("How do I decrease overfitting for Logistic Regression?")


for topic_idx, topic in enumerate(ldia.components_):
          print(f"\nTopic #{topic_idx + 1}:")
          print("; ".join([words[i]
                           for i in topic.argsort()[:-6:-1]]))


import pandas as pd
from matplotlib import pyplot as plt
import seaborn

seaborn.set_theme('notebook')
seaborn.set_style('whitegrid')


df = pd.read_csv('secret_message_convolved_line_plot.csv', index_col=0)
figsize = (10, 6)
fig = plt.figure(figsize=figsize)
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.xlabel('Time (sec)')
plt.ylabel('Signal amplitude (-1 to +1)')
plt.legend()
plt.show()



import pandas as pd
import numpy as np
# from sklearn.linear_model import LogisticRegression, Lasso
from sklearn.feature_extraction.text import TfidfVectorizer

from nlpia.constants import DATA_DIR

df = pd.read_csv(DATA_DIR / 'baby-names-region.csv.gz')


""" Add a freq column based on the counts, to use as sample_weights in LogisticRegression.fit

>>> df['freq'] = 0.
>>> for y in df['year'].unique():
...     g = df[y == df['year']]
...     tot = g['count'].sum()
...     mask = y == df['year']
...     df['freq'][mask] = df['count'][mask] / df['count'][mask].values.sum()
...     print(y, tot, df['freq'][mask].sum())

>>> df['freq'].sum().round(10)
111.0
>>> df['year'].nunique()
111
"""

df = pd.read_csv('.nlpia2-data/baby-names-region.csv.gz')
df['freq'] = 0.
for y in df['year'].unique():
    g = df[y == df['year']]
    tot = g['count'].sum()
    mask = y == df['year']
    df['freq'][mask] = df['count'][mask] / df['count'][mask].values.sum()
    print(y, tot, df['freq'][mask].sum())
df['freq'].sum().round(10)
df['year'].nunique()
df.to_csv('.nlpia2-data/baby-names-region.csv.gz', index=False)

# verify minimial loss of precision during small value sum:
pd.read_csv('.nlpia2-data/baby-names-region.csv.gz')
for y in df['year'].unique():
    g = df[y == df['year']]
    tot = g['count'].sum()
    mask = y == df['year']
    print(y, tot, df['freq'][mask].sum())


df = pd.read_csv('/home/hobs/.nlpia2-data/baby-names-region.csv.gz')

# df = df.sample(1_000_000, random_state=1989)
np.random.seed(451)
istrain = np.random.rand(len(df)) < .9

vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
vectorizer.fit(df['name'][istrain])
vecs = vectorizer.transform(df['name'])



import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer

REPO_URL = 'https://gitlab.com/tangibleai/qary/-/raw/master'
FAQ_DIR = 'src/qary/data/faq'
FAQ_FILENAME = 'short-faqs.csv'
DS_FAQ_URL = '/'.join([REPO_URL, FAQ_DIR, FAQ_FILENAME])

df = pd.read_csv(DS_FAQ_URL)

vectorizer = TfidfVectorizer()
vectorizer.fit(df['question'])
# vectorize all the questions/answers in qa_dataset
tfidfvectors_sparse = vectorizer.transform(df['question'])
tfidfvectors = tfidfvectors_sparse.todense()

question = "What's overfitting a model?"
question_vector = vectorizer.transform([question]).todense()
idx = question_vector.dot(tfidfvectors.T).argmax()
# 51

print(
    f"Your question:\n  {question}\n\n"
    f"Most similar FAQ question:\n  {df['question'][idx]}\n\n"
    f"Answer to that FAQ question:\n  {df['answer'][idx]}\n\n"
)
# Your question:
#   What's overfitting a model?

# Most similar FAQ question:
#   What is overfitting?

# Answer to that FAQ question:
#   When your test set accuracy is significantly lower than your training set accuracy?


question = 'What causes a model to overfit?'
question_vector = vectorizer.transform([question]).todense()
idx = question_vector.dot(tfidfvectors.T).argmax()
# 35
print(
    f"Your question:\n  {question}\n\n"
    f"Most similar FAQ question:\n  {df['question'][idx]}\n\n"
    f"Answer to that FAQ question:\n  {df['answer'][idx]}\n\n"
)
# Your question:
#   What causes a model to overfit?

# Most similar FAQ question:
#   Is a `LogisticRegression` more likely or less likely to overfit to your data when compared to a `DecisionTree`?

# Answer to that FAQ question:
#   A `LogisticRegression` will be less likely to overfit than a `DecisionTree` in most situations.

# To improve recall, a stemmer should be added to the TfidfVectorizer


def tokenizer(text, nlp=None):
    """
    >>> spacy.cli.download('en_core_web_sm')
    >>> nlp = spacy.load('en_core_web_sm')
    >>> tokenizer("What causes overfit?", nlp)
    ['what', 'causes', 'overfit', '?']
    >>> tokenizer("What causes overfitting?", nlp)
    ['what', 'cause', 'overfitte', '?']

    >>> spacy.cli.download('en_core_web_md')
    >>> nlp = spacy.load('en_core_web_md')
    >>> tokenizer("What causes overfitting?", nlp)
    ['what', 'cause', 'overfitting', '?']
    >>> tokenizer("What causes overfit?", nlp)
    ['what', 'cause', 'overfit', '?']
    >>> tokenizer("What causes running?", nlp)
    ['what', 'causes', 'run', '?']

    Lemmatization will handle causes/cause, but not hitting/hit,
    even if you use the large model.

    >>> # spacy.cli.download('en_core_web_lg')
    >>> nlp = spacy.load('en_core_web_lg')
    >>> tokenizer("What causes overhitting?", nlp=nlp)
    ['what', 'cause', 'overhitting', '?']
    # >>> spacy.cli.download('en_core_web_md')
    >>> tokenizer('What causes overfitting?', nlp=nlp)
    ['what', 'cause', 'overfitting', '?']
    >>> tokenizer('What causes over-fitting?', nlp=nlp)
    ['what', 'cause', 'over', '-', 'fitting', '?']
    >>> tokenizer('What causes over fitting?', nlp=nlp)
    ['what', 'cause', 'over', 'fitting', '?']
    >>> tokenizer('What causes over hitting?', nlp=nlp)
    ['what', 'cause', 'over', 'hit', '?']
    >>> tokenizer('What causes over-hitting?', nlp=nlp)
    ['what', 'cause', 'over', '-', 'hit', '?']
    >>> tokenizer('What causes overhitting?', nlp=nlp)
    ['what', 'cause', 'overhitting', '?']
    """
    return [text.lemma_ for text in nlp(txt)]



import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer

REPO_URL = 'https://gitlab.com/tangibleai/qary/-/raw/master'
FAQ_DIR = 'src/qary/data/faq'
FAQ_FILENAME = 'short-faqs.csv'
DS_FAQ_URL = '/'.join([REPO_URL, FAQ_DIR, FAQ_FILENAME])

df = pd.read_csv(DS_FAQ_URL)

vectorizer = TfidfVectorizer()
vectorizer.fit(df['question'])
# vectorize all the questions/answers in qa_dataset
tfidfvectors_sparse = vectorizer.transform(df['question'])
tfidfvectors = tfidfvectors_sparse.todense()

question = "How do I decrease overfitting for Logistic Regression?"
question_vector = vectorizer.transform([question]).todense()
idx = question_vector.dot(tfidfvectors.T).argmax()
# 51

print(
    f"Your question:\n  {question}\n\n"
    f"Most similar FAQ question:\n  {df['question'][idx]}\n\n"
    f"Answer to that FAQ question:\n  {df['answer'][idx]}\n\n"
)
# Your question:
#   What's overfitting a model?

# Most similar FAQ question:
#   What is overfitting?

# Answer to that FAQ question:
#   When your test set accuracy is significantly lower than your training set accuracy?


question = 'What causes a model to overfit?'
question_vector = vectorizer.transform([question]).todense()
idx = question_vector.dot(tfidfvectors.T).argmax()
# 35
print(
    f"Your question:\n  {question}\n\n"
    f"Most similar FAQ question:\n  {df['question'][idx]}\n\n"
    f"Answer to that FAQ question:\n  {df['answer'][idx]}\n\n"
)
# Your question:
#   What causes a model to overfit?

# Most similar FAQ question:
#   Is a `LogisticRegression` more likely or less likely to overfit to your data when compared to a `DecisionTree`?

# Answer to that FAQ question:
#   A `LogisticRegression` will be less likely to overfit than a `DecisionTree` in most situations.

# To improve recall, a stemmer should be added to the TfidfVectorizer



""" word2vec trained on Wikipedia text with pytorch
## Resources

```yaml
-
  citation: Mikolov et al (2013) ARXIV Computation and Language
  url: https://arxiv.org/abs/1301.3781)
-
  citation: Musashi (Jacobs-) Harukawa (21 Oct 2021) Personal Blog
  description: Word2Vec from scratch on tweet hate speech
  url: https://muhark.github.io/python/ml/nlp/2021/10/21/word2vec-from-scratch.html
-
  citation: Garg et al (2018), PNAS
  description: shifting stereotypes over 100 yrs
  url: https://www.pnas.org/content/pnas/115/16/E3635.full.pdf
-
  citation: Rodman (2019), Political Analysis
  description: meaning of "equality" in US
  url: https://static1.squarespace.com/static/5ca7d04ea09a7e68ba44e707/t/5cda219af4e1fc94236bc0cf/1557799325771/Diachronic_Word_Vectors___Political_Analysis_Final_Version.pdf # noqa
-
  citation: Rodriguez and Spirling (2021), Journal of Politics
  description: word embeddings for social science
  url: https://www.journals.uchicago.edu/doi/10.1086/715162
```
"""
from collections import Counter
from itertools import chain
import re

import numpy as np
import pandas as pd
import torch
from torch import nn
from torchtext import datasets
from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm

MIN_FREQ = 10  # ignore words that appear fewer than 10 times
WINDOW_WIDTH = 3

# ~1.8M lines of wikipedia text: https://pytorch.org/text/stable/datasets.html#wikitext-103
# dset = datasets.Wikitext103()

# ~40k lines (paragraphs) of Wikipedia text: https://pytorch.org/text/stable/datasets.html#wikitext-2
# Short and empty lines (nonsentences) are not filtered out
dsets = datasets.WikiText2()

pattern = r"\w+(?:\'\w+)?|[A-Z][a-z0-9]+|[^\w\s]+"
re_pattern = re.compile(pattern)


# chain all 3 sample sets in dset[0] (train), dset[1] (validation), dset[2] (test)
tokenized_lines = map(re_pattern.findall, chain(*dsets))

counts = Counter(chain(*tokenized_lines))
print(len(counts))
# 33153

counts = {k: v for k, v in counts.items() if v > MIN_FREQ}  # ignore rare terms
vocab = list(counts.keys())
# pd.Series()
vocab_size = len(vocab)
id2tok = dict(enumerate(vocab))
tok2id = {token: id for id, token in id2tok.items()}

len(counts)
# 15047


def delete_oov(dset):
    return [[tok for tok in line if tok in vocab] for line in dset]


def neighbor_pairs(tokens, window_width=WINDOW_WIDTH):
    """ Mikolov called these skip grams - pairs of words that are within window_width of each other

    >>> neighbor_pairs('Every good boy does fine .'.split(), 3)
    [
     (Every, good),
     (Every, boy),
     (Every, does),
     (good, boy),
     (good, does),
     (good, fine),
    ...
    """
    pairs = []
    for i, wd in enumerate(tokens):
        target = tok2id[wd]
        window = [
            i + j for j in
            range(-window_width, window_width + 1, 1)
            if (i + j >= 0)
            & (i + j < len(tokens))
            & (j != 0)
        ]

        pairs.extend([(target, tok2id[tokens[w]]) for w in window])
    # huggingface datasets are dictionaries for every text element
    # line['moving_window'] = pairs
    # return line
    return pairs


###################################################
# dataloader


class Word2VecDataset(Dataset):
    """
    Takes a PyTorch Dataset as an input, to be used for a Word2Vec dataloader.
    """

    # FIXME: https://muhark.github.io/python/ml/nlp/2021/10/21/word2vec-from-scratch.html
    # FIXME: original blog used dataset['moving_window'] in place of pairs
    def __init__(self, pairs, vocab_size):  # , window_width=WINDOW_WIDTH):
        self.dataset = pairs
        self.vocab_size = vocab_size
        self.data = np.array([i for s in pairs for i in s])

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]


BATCH_SIZE = 2**14
N_LOADER_PROCS = 10


dsets = datasets.WikiText2()
dataloader = {}
# Convert pytorch raw text datasets (3-tuple) to a dict with 3 key-value pairs for train/val/test
dataset = {'train': list(dsets[0]), 'validation': list(dsets[1]), 'test': list(dsets[2])}

for key in dataset.keys():
    dataloader = {
        key: DataLoader(
            Word2VecDataset(dataset[key], vocab_size=vocab_size),
            batch_size=BATCH_SIZE,
            shuffle=True,
            num_workers=N_LOADER_PROCS
        )
    }

# dataloader
###################################################


def one_hot_encode(token_id, vocab_size):
    vec = torch.zeros(vocab_size).float()
    vec[token_id] = 1.0
    return vec


ohe = one_hot_encode(token_id=3, vocab_size=10)
linear_layer = nn.Linear(vocab_size, 1, bias=False)

# Set edge weights from 0 to 9 for easy reference
with torch.no_grad():
    linear_layer.weight = nn.Parameter(
        torch.arange(10, dtype=torch.float).reshape(linear_layer.weight.shape))

print(linear_layer.weight)
# Parameter containing:
# tensor([[0., 1., 2., 3., 4., 5., 6., 7., 8., 9.]], requires_grad=True)

print(linear_layer(ohe))
# tensor([3.], grad_fn=<SqueezeBackward3>)


embedding_layer = nn.Embedding(vocab_size, 1)

with torch.no_grad():
    embedding_layer.weight = nn.Parameter(
        torch.arange(10, dtype=torch.float
                     ).reshape(embedding_layer.weight.shape))

print(embedding_layer.weight)
print(embedding_layer(torch.tensor(input)))


class Word2Vec(nn.Module):
    def __init__(self, vocab_size, embedding_size):
        super().__init__()
        self.embed = nn.Embedding(vocab_size, embedding_size)
        self.expand = nn.Linear(embedding_size, vocab_size, bias=False)

    def forward(self, input):
        # Encode input to lower-dimensional representation
        hidden = self.embed(input)
        # Expand hidden layer to predictions
        logits = self.expand(hidden)
        return logits


# Instantiate the model
EMBED_SIZE = 100  # Mikalov found that 300-D word vectors worked well
model = Word2Vec(vocab_size, EMBED_SIZE)

# Relevant if you have a GPU:
device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
model.to(device)

# Define training parameters
LR = 3e-4
EPOCHS = 10
loss_fn = nn.CrossEntropyLoss()
optimizer = torch.optim.AdamW(model.parameters(), lr=LR)


pbar = tqdm(range(EPOCHS * len(dataloader['train'])))
running_loss = []
for epoch in range(EPOCHS):
    epoch_loss = 0
    for center, context in dataloader['train']:
        center, context = center.to(device), context.to(device)
        optimizer.zero_grad()
        logits = model(input=context)
        loss = loss_fn(logits, center)
        epoch_loss += loss.item()
        loss.backward()
        optimizer.step()
        pbar.update(1)
    epoch_loss /= len(dataloader['train'])
    running_loss.append(epoch_loss)



import numpy as np
np.random.seed(451)  # for reproducibility

from keras.preprocessing import sequence
from keras.models import Sequential
from keras.layers import Dense, Activation, Embedding
from keras.layers import LSTM
from keras.datasets import imdb

max_features = 20000
maxlen = 80  # cut texts after this number of words (among top max_features most common words)
batch_size = 32

print('Loading data...')
(X_train, y_train), (X_test, y_test) = imdb.load_data(nb_words=max_features)
print(len(X_train), 'train sequences')
print(len(X_test), 'test sequences')

print('Pad sequences (samples x time)')
X_train = sequence.pad_sequences(X_train, maxlen=maxlen)
X_test = sequence.pad_sequences(X_test, maxlen=maxlen)
print('X_train shape:', X_train.shape)
print('X_test shape:', X_test.shape)

print('Build model...')
model = Sequential()
model.add(Embedding(max_features, 128, dropout=0.2))
model.add(LSTM(128, dropout_W=0.2, dropout_U=0.2))  # try using a GRU instead, for fun
model.add(Dense(1))
model.add(Activation('sigmoid'))

# try using different optimizers and different optimizer configs
model.compile(loss='binary_crossentropy',
              optimizer='adam',
              metrics=['accuracy'])

print('Train...')
model.fit(X_train, y_train, batch_size=batch_size, nb_epoch=15,
          validation_data=(X_test, y_test))
score, acc = model.evaluate(X_test, y_test,
                            batch_size=batch_size)
print('Test score:', score)
print('Test accuracy:', acc)
